import numpy as np
import matplotlib.pyplot as plt
from matplotlib import gridspec
import seaborn as sn
from mpl_toolkits.mplot3d import Axes3D
import matplotlib.colors as colors
from mpl_toolkits.mplot3d.art3d import Poly3DCollection
from concurrent.futures import ProcessPoolExecutor as PPE
from concurrent.futures import as_completed
from skimage import measure
from time import time
from tqdm import tqdm
from pickle import loads
#import pandas as pd
import vtk
import plotly.graph_objects as go

from ..tests.bumpy_sphere import bumpy_sphere

def mpu_meshgrid(mpu_object,buf=1.25,res=10,workers=1,k=None,plane_axis=None,plane_value=None,marching=False,gradient=False,record=False,global_f=False):
    dims = []
    if not marching:
        for d in range(mpu_object.ddim):
            if d is plane_axis:
                continue
            dims.append(np.linspace(mpu_object.dim_range[d*2]-buf,
                                    mpu_object.dim_range[d*2+1]+buf,res))
    else:
        res = res*1j
        x,y,z = np.ogrid[mpu_object.dim_range[0]-buf:mpu_object.dim_range[1]+buf:res,
                         mpu_object.dim_range[2]-buf:mpu_object.dim_range[3]+buf:res,
                         mpu_object.dim_range[4]-buf:mpu_object.dim_range[5]+buf:res]
        results = mpu_object.function_marching(x,y,z)
        return (x,y,z),results
    DIMS = np.meshgrid(*dims)
    DIMSf = []
    for d in range(len(DIMS)):
        DIMSf.append(DIMS[d].flatten())
    value = []
    if plane_value is not None:
        plane_values = [plane_value]*len(DIMSf[0])
        DIMSf.insert(plane_axis,plane_values)
        plane = np.array(plane_values).reshape(DIMS[0].shape)
        DIMS.insert(plane_axis,plane)
    if workers > 1:
        chunksize = len(DIMSf[0])//workers
        number_of_chunks = len(DIMSf[0])//chunksize
        executor = PPE(max_workers=workers)
        if global_f:
            pass
        elif k is not None:
            k = [k]*len(DIMSf[0])
            DIMSf.append(k)
        else:
            k = [len(mpu_object.patches)]*len(DIMSf[0])
            DIMSf.append(k)
        if not gradient:
            result_generator  = executor.map(loads(mpu_object.pickled_DD[0]),zip(*DIMSf),chunksize=chunksize)
        else:
            result_generator  = executor.map(loads(mpu_object.pickled_DD[1]),zip(*DIMSf),chunksize=chunksize)
        executor.shutdown(wait=True)
        if not gradient:
            results = []
            for result in result_generator:
                results.append(result)
            results = np.array(results)
            results = results.reshape(DIMS[0].shape)
        else:
            u = []
            v = []
            w = []
            for result in result_generator:
                u.append(result[0])
                v.append(result[1])
                w.append(result[2])
            u = np.array(u)
            u = u.reshape(DIMS[0].shape)
            v = np.array(v)
            v = u.reshape(DIMS[0].shape)
            w = np.array(w)
            w = u.reshape(DIMS[0].shape)
            results = [u,v,w]
    else:
        time_data = []
        results = []
        if global_f:
            pass
        elif k is not None:
            k = [k]*len(DIMSf[0])
            DIMSf.append(k)
        else:
            k = [len(mpu_object.patches)]*len(DIMSf[0])
            DIMSf.append(k)
        if not gradient:
            for i in zip(*DIMSf):
                start = time()
                results.append(mpu_object.DD[0](i))
                stop = time() - start
                time_data.append(stop)
            results = np.array(results)
            results = results.reshape(DIMS[0].shape)
            if record:
                results = time_data
        else:
            u = []
            v = []
            w = []
            for i in zip(*DIMSf):
                start = time()
                results = mpu_object.DD[1](i)
                stop = time() - start
                time_data.append(stop)
                u.append(results[0])
                v.append(results[1])
                w.append(results[2])
            u = np.array(u)
            u = u.reshape(DIMS[0].shape)
            v = np.array(v)
            v = v.reshape(DIMS[0].shape)
            w = np.array(w)
            w = w.reshape(DIMS[0].shape)
            results = [u,v,w]
            if record:
                results = time_data
    return DIMS,results

def plot_volume(mpu_object,resolution=20,workers=1,cmin=-1,cmax=0,surface_count=14,k=None,global_f=False,show_points=False):
    """
    This function accepts a perfusion domain object instance and returns a figure
    object for rendering the domain for visualization.

    Parameters
    ----------
              mpu_object: svcco.surface object instance

              resolution: int (default: 20)
                          The number of points along each axis to sample when
                          determining the contours of the implicit function
                          representing the domain.

                          The total number of sampled points scales cubically
                          to the resolution size. (i.e. resolution=20;
                          total_points=8000)

              workers: int (default: 1)
                          The number of processes to split the function evaluations
                          among. Becuase volume rendering can easily require tens
                          of thousands of function evaluations
    """
    DIMS,results = mpu_meshgrid(mpu_object,res=resolution,workers=workers,k=k,global_f=global_f)
    if show_points:
        fig = go.Figure(data=(go.Isosurface(
                        x=DIMS[0].flatten(),
                        y=DIMS[1].flatten(),
                        z=DIMS[2].flatten(),
                        value=results.flatten(),
                        opacity=0.1,
                        isomin=cmin,
                        isomax=cmax,
                        surface_count=surface_count),go.Scatter3d(x = mpu_object.points[:,0],y=mpu_object.points[:,1],z=mpu_object.points[:,2])))
    else:
        fig = go.Figure(data=(go.Isosurface(
                        x=DIMS[0].flatten(),
                        y=DIMS[1].flatten(),
                        z=DIMS[2].flatten(),
                        value=results.flatten(),
                        opacity=0.1,
                        isomin=cmin,
                        isomax=cmax,
                        surface_count=surface_count)))
    fig.show()
    return fig

def plot_volume_individual(func,x_range,y_range,z_range,buf=1.25,res=20,surface_count=15,cmin=-1,cmax=0):
    res = res*1j
    x,y,z = np.ogrid[x_range[0]-buf:x_range[1]+buf:res,
                     y_range[0]-buf:y_range[1]+buf:res,
                     z_range[0]-buf:z_range[1]+buf:res]
    results = func(x,y,z)
    fig = go.Figure(data=(go.Isosurface(
                    x=x.flatten(),
                    y=y.flatten(),
                    z=z.flatten(),
                    value=results.flatten(),
                    opacity=0.1,
                    isomin=cmin,
                    isomax=cmax,
                    surface_count=surface_count)))
    fig.show()

def plot_slice_error(mpu_object,resolution=20,k=None,workers=1,plane_axis=2,plane_value=0,unit_scale='mm'):
    DIMS,results_exact = mpu_meshgrid(mpu_object,res=resolution,workers=workers,
                                      k=len(mpu_object.patches),plane_axis=plane_axis,
                                      plane_value=plane_value)
    DIMS,results_approximate = mpu_meshgrid(mpu_object,res=resolution,workers=workers,
                                      k=k,plane_axis=plane_axis,plane_value=plane_value)
    DIMSv,voxels = mpu_meshgrid(mpu_object,res=resolution,workers=workers,
                                k=len(mpu_object.patches),marching=True)
    verts, faces, normals, values = measure.marching_cubes(voxels,level=0)
    verts *= np.array([np.diff(ar.flat)[0] for ar in [DIMSv[0],DIMSv[1],DIMSv[2]]])
    verts += np.array([DIMSv[0].min(),DIMSv[1].min(),DIMSv[2].min()])

    error = results_exact - results_approximate
    error_min = error.min()
    error_max = error.max()
    error_divide = max(abs(error_min),abs(error_max))
    error = error/(error_divide)
    colormap = plt.cm.RdBu
    main_axes = list(range(len(DIMS)))
    main_axes.remove(plane_axis)
    axes_labels = ['x ({})'.format(unit_scale),'y ({})'.format(unit_scale),'z ({})'.format(unit_scale)]
    axes_labels_all = ['x ({})'.format(unit_scale),'y ({})'.format(unit_scale),'z ({})'.format(unit_scale)]
    axes_labels.remove(axes_labels[plane_axis])
    fig = plt.figure(figsize=(12,6))
    spec = gridspec.GridSpec(ncols=4, nrows=1,
                             width_ratios=[1, 1, 1, 1.5])
    ax1 = fig.add_subplot(spec[0])
    ax2 = fig.add_subplot(spec[1])
    ax3 = fig.add_subplot(spec[2])
    ax4 = fig.add_subplot(spec[3],projection='3d')

    cax1 = ax1.contourf(DIMS[main_axes[0]],DIMS[main_axes[1]],results_exact,norm=colors.TwoSlopeNorm(0))
    ax1.contour(DIMS[main_axes[0]],DIMS[main_axes[1]],results_exact,levels=[0],linewidths=(2,))
    #ax1.set_title('Exact Interpolation \n {} Patches'.format(len(mpu_object.patches)))
    ax1.set_xlabel(axes_labels[0])
    ax1.set_ylabel(axes_labels[1])

    cax2 = ax2.contourf(DIMS[main_axes[0]],DIMS[main_axes[1]],results_approximate,norm=colors.TwoSlopeNorm(0))
    ax2.contour(DIMS[main_axes[0]],DIMS[main_axes[1]],results_approximate,levels=[0],linewidths=(2,))
    #ax2.set_title('Approximate Interpolation \n {} Patches'.format(k))
    ax2.set_xlabel(axes_labels[0])

    cax3 = ax3.contourf(DIMS[main_axes[0]],DIMS[main_axes[1]],error,cmap=colormap,norm=colors.TwoSlopeNorm(0))

    #ax3.set_title('Relative Truncation Error')
    ax3.set_xlabel(axes_labels[0])

    mesh = Poly3DCollection(verts[faces])
    mesh.set_edgecolor('k')
    ax4.add_collection3d(mesh)
    ax4.plot_surface(DIMS[0],DIMS[1],DIMS[2],alpha=0.2,color='r')
    ax4.set_xlim(DIMSv[0].min(),DIMSv[0].max())
    ax4.set_ylim(DIMSv[1].min(),DIMSv[1].max())
    ax4.set_zlim(DIMSv[2].min(),DIMSv[2].max())
    ax4.set_xlabel(axes_labels_all[0])
    ax4.set_ylabel(axes_labels_all[1])
    ax4.set_zlabel(axes_labels_all[2])
    plt.draw()
    p0 = ax1.get_position().get_points().flatten()
    p1 = ax2.get_position().get_points().flatten()
    p2 = ax3.get_position().get_points().flatten()
    ax_cbar = fig.add_axes([p0[0], 0.1, p1[2]-p0[0], 0.05])
    ax_cbar.set_in_layout(True)
    cbar1 = plt.colorbar(cax1, cax=ax_cbar, orientation='horizontal', pad=0.25)
    labels = cbar1.ax.get_xticklabels()
    cbar1.ax.set_xticklabels(labels, rotation=40)


    ax_cbar1 = fig.add_axes([p2[0], 0.1, p2[2]-p2[0], 0.05])
    ax_cbar1.set_in_layout(True)
    cbar2 = plt.colorbar(cax3, cax=ax_cbar1, orientation='horizontal', pad=0.25)
    labels = cbar2.ax.get_xticklabels()
    cbar2.ax.set_xticklabels(labels, rotation=40)
    plt.subplots_adjust(bottom=0.25)

    plt.show()

def level_contour(mpu_object,resolution=20,k=2,buf=1.25,level=0,workers=1,plane_axis=2,plane_value=0.5,unit_scale='mm',name="untitled"):
    if plane_axis == 0:
        plane_value = (mpu_object.x_range[1]-mpu_object.x_range[0])*plane_value + mpu_object.x_range[0]
    elif plane_axis == 1:
        plane_value = (mpu_object.y_range[1]-mpu_object.y_range[0])*plane_value + mpu_object.y_range[0]
    elif plane_axis == 2:
        plane_value = (mpu_object.z_range[1]-mpu_object.z_range[0])*plane_value + mpu_object.z_range[0]
    DIMS,results_exact = mpu_meshgrid(mpu_object,res=resolution,workers=workers,
                                      k=k,plane_axis=plane_axis,
                                      plane_value=plane_value,buf=buf)
    colormap = plt.cm.RdBu
    main_axes = list(range(len(DIMS)))
    main_axes.remove(plane_axis)
    fig,ax = plt.subplots()
    CS = ax.contour(DIMS[main_axes[0]],DIMS[main_axes[1]],results_exact,levels=[level],linewidths=(2,))
    #CS.collections[0].set_label(r"$\Gamma_{" +str(len(mpu_object.patches))+ r"}$")
    return CS

def plot_mpu_boundaries(mpu_object,resolution=20,k=[2],buf=1.25,level=0,workers=1,plane_axis=2,plane_value=0.5,test_size=100,unit_scale='mm',name="untitled"):
    if plane_axis == 0:
        plane_value = (mpu_object.x_range[1]-mpu_object.x_range[0])*plane_value + mpu_object.x_range[0]
    elif plane_axis == 1:
        plane_value = (mpu_object.y_range[1]-mpu_object.y_range[0])*plane_value + mpu_object.y_range[0]
    elif plane_axis == 2:
        plane_value = (mpu_object.z_range[1]-mpu_object.z_range[0])*plane_value + mpu_object.z_range[0]
    DIMS,results_exact = mpu_meshgrid(mpu_object,res=resolution,workers=workers,
                                      k=len(mpu_object.patches),plane_axis=plane_axis,
                                      plane_value=plane_value,buf=buf)
    APPROX = []
    for ki in k:
        DIMS,approximate = mpu_meshgrid(mpu_object,res=resolution,workers=workers,
                                          k=ki,plane_axis=plane_axis,plane_value=plane_value)
        APPROX.append(approximate)

    colormap = plt.cm.RdBu
    main_axes = list(range(len(DIMS)))
    main_axes.remove(plane_axis)
    axes_labels = ['x ({})'.format(unit_scale),'y ({})'.format(unit_scale),'z ({})'.format(unit_scale)]
    axes_labels_all = ['x ({})'.format(unit_scale),'y ({})'.format(unit_scale),'z ({})'.format(unit_scale)]
    axes_labels.remove(axes_labels[plane_axis])
    fig,ax = plt.subplots()
    #spec = gridspec.GridSpec(ncols=4, nrows=1,
    #                         width_ratios=[1, 1, 1, 1.5])
    #ax1 = fig.add_subplot(spec[0])
    #ax2 = fig.add_subplot(spec[1])
    #ax3 = fig.add_subplot(spec[2])
    #ax4 = fig.add_subplot(spec[3],projection='3d')
    positions = []
    cax = ax.contourf(DIMS[main_axes[0]],DIMS[main_axes[1]],results_exact,norm=colors.TwoSlopeNorm(0),cmap="RdBu_r",alpha=0.5)
    CS = ax.contour(DIMS[main_axes[0]],DIMS[main_axes[1]],results_exact,levels=[level],linewidths=(2,))
    CS.collections[0].set_label(r"$\Gamma_{" +str(len(mpu_object.patches))+ r"}$")
    #fmt = {CS.levels[0]:r"$\Gamma_{" +str(len(mpu_object.patches)) + r"}$"}
    #ax.clabel(CS,fmt=fmt,fontsize=10)
    for c,ap in enumerate(APPROX):
        tmp = ax.contour(DIMS[main_axes[0]],DIMS[main_axes[1]],ap,levels=[level],linewidths=(2,),linestyles='--')
        tmp.collections[0].set_label(r"$\Gamma_{" +str(k[c])+ r"}$")
        #fmt = {tmp.levels[0]:r"$\Gamma_{" +str(k[c])+ r"}$"}
        #ax.clabel(tmp,fmt=fmt,fontsize=10)
    plt.legend()
    pts = np.random.random((test_size,3))
    xr = (np.min(DIMS[0].flatten()),np.max(DIMS[0].flatten()))
    yr = (np.min(DIMS[1].flatten()),np.max(DIMS[1].flatten()))
    zr = (np.min(DIMS[2].flatten()),np.max(DIMS[2].flatten()))
    pts[:,0] = (pts[:,0]*2-1)*(xr[1] - xr[0])/2 + (mpu_object.x_range[0] + mpu_object.x_range[1])/2
    pts[:,1] = (pts[:,1]*2-1)*(yr[1] - yr[0])/2 + (mpu_object.y_range[0] + mpu_object.y_range[1])/2
    pts[:,2] = (pts[:,2]*2-1)*(zr[1] - zr[0])/2 + (mpu_object.z_range[0] + mpu_object.z_range[1])/2
    pts[:,plane_axis] = plane_value
    interior = []
    exterior = []
    interior_wrong = []
    exterior_wrong = []
    for i in range(pts.shape[0]):
        if mpu_object.within(pts[i,0],pts[i,1],pts[i,2],len(mpu_object.patches),level=level)[0]:
            if mpu_object.within(pts[i,0],pts[i,1],pts[i,2],k[0],level=level)[0]:
                interior.append(pts[i,:].tolist())
            else:
                exterior_wrong.append(pts[i,:].tolist())
        else:
            if not mpu_object.within(pts[i,0],pts[i,1],pts[i,2],k[0],level=level)[0]:
                exterior.append(pts[i,:].tolist())
            else:
                interior_wrong.append(pts[i,:].tolist())
    interior = np.array(interior)
    exterior = np.array(exterior)
    interior_wrong = np.array(interior_wrong)
    exterior_wrong = np.array(exterior_wrong)
    if interior.shape[0] > 0:
        if interior.shape[0] > 100:
            stop = 99
        else:
            stop = None
        ax.scatter(interior[:stop,main_axes[0]],interior[:stop,main_axes[1]],c='blue',alpha=0.7)
    if exterior.shape[0] > 0:
        if exterior.shape[0] > 100:
            stop = 99
        else:
            stop = None
        ax.scatter(exterior[:stop,main_axes[0]],exterior[:stop,main_axes[1]],c='red',alpha=0.7)
    if exterior_wrong.shape[0] > 0:
        if exterior_wrong.shape[0] > 100:
            stop = 99
        else:
            stop = None
        ax.scatter(exterior_wrong[:stop,main_axes[0]],exterior_wrong[:stop,main_axes[1]],c='red',alpha=0.7,marker='x')
    if interior_wrong.shape[0] > 0:
        if interior_wrong.shape[0] > 100:
            stop = 99
        else:
            stop = None
        ax.scatter(interior_wrong[:stop,main_axes[0]],interior_wrong[:stop,main_axes[1]],c='blue',alpha=0.7,marker='x')
    plt.savefig(name+'.svg', format="svg", bbox_inches='tight')
    return interior.shape[0],exterior.shape[0],interior_wrong.shape[0],exterior_wrong.shape[0],test_size

def plot_gradient(mpu_object,workers=1,k=None,resolution=20,plane_axis=2,plane_value=0,
                  contour_overlay=True,unit_scale='mm'):
    DIMS,gradient = mpu_meshgrid(mpu_object,res=resolution,workers=workers,
                                 plane_axis=plane_axis,plane_value=plane_value,
                                 gradient=True)
    DIMS,gradient_approximate = mpu_meshgrid(mpu_object,res=resolution,workers=workers,
                                             plane_axis=plane_axis,plane_value=plane_value,
                                             k=k,gradient=True)
    if contour_overlay:
        DIMS_c,result = mpu_meshgrid(mpu_object,res=resolution,workers=workers,
                                     plane_axis=plane_axis,plane_value=plane_value)
    main_axes = list(range(len(DIMS)))
    main_axes.remove(main_axes[plane_axis]) #use del?
    fig = plt.figure(figsize=(6,5))
    ax = fig.add_subplot(111)

    if contour_overlay:
        cax = ax.contourf(DIMS[main_axes[0]],DIMS[main_axes[1]],result,norm=colors.TwoSlopeNorm(0))
        p0 = ax.get_position().get_points().flatten()
        ax_cbar = fig.add_axes([p0[0], 0.1, p0[2]-p0[0], 0.05])
        ax_cbar.set_in_layout(True)
        cbar1 = plt.colorbar(cax, cax=ax_cbar, orientation='horizontal', pad=0.25)
        labels = cbar1.ax.get_xticklabels()
        cbar1.ax.set_xticklabels(labels, rotation=40)

    ax.quiver(DIMS[main_axes[0]],DIMS[main_axes[1]],
              gradient[main_axes[0]],gradient[main_axes[1]])

    ax.quiver(DIMS[main_axes[0]],DIMS[main_axes[1]],
              gradient_approximate[main_axes[0]],gradient_approximate[main_axes[1]],color='r')
    axes_labels = ['x ({})'.format(unit_scale),'y ({})'.format(unit_scale),'z ({})'.format(unit_scale)]
    axes_labels.remove(axes_labels[plane_axis])
    ax.set_xlabel(axes_labels[0])
    ax.set_ylabel(axes_labels[1])
    plt.subplots_adjust(bottom=0.25)
    plt.show()

def plot_gradient_error(mpu_object,workers=1,k=None,resolution=20,plane_axis=2,plane_value=0,
                  contour_overlay=True,unit_scale='mm'):
    DIMS,gradient = mpu_meshgrid(mpu_object,res=resolution,workers=workers,
                                 plane_axis=plane_axis,plane_value=plane_value,
                                 gradient=True)
    DIMS,approximate = mpu_meshgrid(mpu_object,res=resolution,workers=workers,
                                    plane_axis=plane_axis,plane_value=plane_value,
                                    k=k,gradient=True)
    DIMS_c,result = mpu_meshgrid(mpu_object,res=resolution,workers=workers,
                                 plane_axis=plane_axis,plane_value=plane_value)
    gradient = [gradient[0].flatten(),gradient[1].flatten(),gradient[2].flatten()]
    approximate = [approximate[0].flatten(),approximate[1].flatten(),approximate[2].flatten()]
    alignment_error = []
    magnitude_error = []
    for i in range(len(gradient[0])):
        grad_mag = (gradient[0][i]**2+gradient[1][i]**2+gradient[2][i]**2)**(1/2)
        approx_mag = (approximate[0][i]**2+approximate[1][i]**2+approximate[2][i]**2)**(1/2)
        alignment_error.append((gradient[0][i]/grad_mag)*(approximate[0][i]/approx_mag)+(gradient[1][i]/grad_mag)*(approximate[1][i]/approx_mag)+
                               (gradient[2][i]/grad_mag)*(approximate[2][i]/approx_mag))
        magnitude_error.append(grad_mag-approx_mag)
    fig = plt.figure()
    ax1 = fig.add_subplot(121)
    ax1.scatter(result.flatten(),alignment_error)
    ax1.set_xlabel('Contour Layer')
    ax1.set_ylabel('Alignment Difference')
    ax2 = fig.add_subplot(122)
    ax2.scatter(result.flatten(),magnitude_error)
    ax2.set_xlabel('Contour Layer')
    ax2.set_ylabel('Magnitude Difference')
    plt.show()

def plot_error(mpu_object,workers=1,k=None,resolution=20,plane_axis=2,plane_value=0):
    DIMS,result = mpu_meshgrid(mpu_object,workers=workers,res=resolution,plane_axis=plane_axis,
                               plane_value=plane_value)
    DIMS,approx = mpu_meshgrid(mpu_object,workers=workers,k=k,res=resolution,plane_axis=plane_axis,
                               plane_value=plane_value)
    fig = plt.figure()
    ax = fig.add_subplot(111)
    main_axes = list(range(len(DIMS)))
    main_axes.remove(main_axes[plane_axis])
    contours = ax.contour(DIMS[main_axes[0]],DIMS[main_axes[1]],result)
    levels = contours.levels[1:-1]
    level_areas = []
    level_error = []
    error_per_area = []
    resultf = result.flatten()
    approxf = approx.flatten()
    l0 = -1
    a0 = 0
    passed_levels = []
    for idx, level in enumerate(levels):
        #contour = contours.collections[idx]
        #if len(contour.get_paths()) == 0:
        #    continue
        #else:
        #    passed_levels.append(level)
        #vs = contour.get_paths()[0].vertices
        #a = 0
        #d0_0,d1_0 = vs[0]
        #for [d0_1,d1_1] in vs[1:]:
        #    dd0 = d0_1 - d0_0
        #    dd1 = d1_1 - d1_0
        #    a += 0.5*(d1_0*dd0 - d0_0*dd1)
        #    d0_0 = d0_1
        #    d1_0 = d1_1
        #level_areas.append(a-a0)
        #a0 = a
        lower_filter = set(np.argwhere(l0<resultf).flatten())
        upper_filter = set(np.argwhere(resultf<level).flatten())
        bounded_filter = list(lower_filter.intersection(upper_filter))
        level_error.append(np.mean(resultf[bounded_filter] - approxf[bounded_filter]))
        l0 = level
        #error_per_area.append(level_error[-1]/level_areas[-1])
    plt.clf()
    plt.plot(levels,level_error)
    plt.show()

def plot_time(mpu_object,resolution=20,k_start=2,k_end=100,gradient=False,plane_axis=None,plane_value=None):
    all_time = []
    mean_time = []
    k_values  = []
    sd_time_values = []
    all_error = []
    mean_error = []
    sd_error = []
    all_k = []
    #d_type = []
    DIMS,exact = mpu_meshgrid(mpu_object,k=None,res=resolution,gradient=gradient,
                                  plane_axis=plane_axis,plane_value=plane_value)
    for i in tqdm(range(k_start,k_end+1),desc='Evaluating Time Grids  '):
        k_values.append(i)
        DIMS,time_data = mpu_meshgrid(mpu_object,k=i,res=resolution,gradient=gradient,
                                      plane_axis=plane_axis,plane_value=plane_value,
                                      record=True)
        DIMS,approx = mpu_meshgrid(mpu_object,k=i,res=resolution,gradient=gradient,
                                      plane_axis=plane_axis,plane_value=plane_value)
        error = abs(exact.flatten() - approx.flatten())
        k_tmp = [i]*len(error)
        #d_type.extend(['error']*len(error))
        #d_type.extend(['time']*len(time_data))
        mean_error.append(np.mean(error))
        sd_error.append(np.std(error))
        all_time.extend(time_data)
        mean_time.append(np.mean(time_data))
        sd_time_values.append(np.std(time_data))
        all_error.extend(error)
        all_k.extend(k_tmp)

    data = pd.DataFrame()
    data['times'] = all_time
    data['k'] = all_k
    data['error'] = all_error
    sn.lineplot(data=data,x='k',y='times')
    ax2 = plt.twinx()
    sn.lineplot(data=data,x='k',y='error',ax=ax2,color='red')
    plt.show()

def time_PU_global(mpu_object):
    start = time()
    mpu_object.solve()
    pu_time = time()-start
    start = time()
    mpu_object.solve(PU=False)
    global_time = time()-start
    return pu_time,global_time

#def pu_v_global(mpu_object,start=10,stop=40):

def show_path(points,points2):
    fig = plt.figure()
    ax  = fig.add_subplot(111,projection='3d')
    ax.scatter3D(points[0,0],points[0,1],points[0,2],color='g')
    ax.scatter3D(points[1:,0],points[1:,1],points[1:,2],color='r')
    ax.scatter3D(points2[:,0],points2[:,1],points2[:,2],color='b')
    plt.show()

def show_mesh(points,points2,points3):
    fig = plt.figure()
    ax  = fig.add_subplot(111,projection='3d')
    ax.scatter3D(points[:,0],points[:,1],points[:,2],color='r')
    ax.scatter3D(points2[:,0],points2[:,1],points2[:,2],color='b')
    ax.scatter3D(points3[:,0],points3[:,1],points3[:,2],color='g')
    plt.show()

def newton(mpu_object):
    f1 = mpu_object.function([1,1,1,None])
    f2 = mpu_object.function([1.1,1,1,None])
    g = np.array(mpu_object.gradient([1,1,1,None]))
    print('F1: {}'.format(f1))
    print('F2: {}'.format(f2))
    print('Ap: {}'.format(f1+np.dot(np.array([0.1,0,0]),g)))

def show_patches(mpu_object):
    patch_sets = [set(tuple([tuple(pp) for pp in p.points.tolist()])) for p in mpu_object.patches]
    unique_points = []
    overlap_points = []
    for patch_idx,patch_i in enumerate(patch_sets):
        patch_i_diff = patch_i.difference({})
        patch_i_overlap = set([])
        for patch_jdx,patch_j in enumerate(patch_sets):
            if patch_idx == patch_jdx:
                continue
            else:
                patch_i_diff = patch_i_diff.difference(patch_j)
                patch_i_overlap.update(patch_i.intersection(patch_j))
            print(len(unique_points))
        unique_points.extend(list(patch_i_diff))
        overlap_points.extend(list(patch_i_overlap))
    unique_points = np.array(unique_points)
    overlap_points = np.array(overlap_points)
    fig = plt.figure()
    ax  = fig.add_subplot(111,projection='3d')
    ax.scatter3D(unique_points[:,0],unique_points[:,1],unique_points[:,2],color='r')
    ax.scatter3D(overlap_points[:,0],overlap_points[:,1],overlap_points[:,2],color='b')
    plt.show()
#get area of contourf for L1 error estimate as a function of "tissue depth" which will basically argue
#that for our area of interest we can ignore the problematic areas of unstable far-field growth

#for gradient error take the angle and magnitude differences and also show them as a function of tissue
# layer position



#!/usr/bin/env python
# coding=utf-8

import argparse
import glob
import os
import csv
import re
import sys
import scipy
import pdb
import vtk
from scipy.interpolate import interp1d

import numpy as np
from collections import defaultdict, OrderedDict

from common import get_dict
from get_database import Database, Post, input_args
from vtk_functions import read_geo, write_geo, collect_arrays, get_all_arrays, ClosestPoints
from get_bc_integrals import get_res_names
from vtk_to_xdmf import write_xdmf
from simulation_io import get_caps_db, collect_results_db, collect_results_db_3d_3d, \
    collect_results_db_3d_3d_spatial

from vtk.util.numpy_support import numpy_to_vtk as n2v
from vtk.util.numpy_support import vtk_to_numpy as v2n

# interpolate 1d to 3d in space and time (allow extrapolation due to round-off errors at bounds)
interp = lambda x_1d, y_1d, x_3d: interp1d(x_1d, y_1d.T, fill_value='extrapolate')(x_3d)

f = 'pressure'
out_path = '/home/pfaller/work/osmsc/cover/'

def main(db, geometries):
    # get post-processing constants
    post = Post()

    for geo in geometries:
        # generate plots
        print(geo)

        # read results
        deform = 'deformable' in db.study
        res, time = collect_results_db(db, geo, post.models, deformable=deform)

        # name of reference solution
        m_ref = '3d_rerun'
        roms = ['0d', '1d']

        # times
        t_ref = time[m_ref]

        res_all = defaultdict(list)

        err = []
        for m_rom in roms:
            # reference time
            t_rom = time[m_rom]

            for br in list(res.keys()):
                # retrieve 3d results
                res_ref = res[br][f][m_ref + '_int_last']

                # map paths to interval [0, 1]
                path_rom = res[br][m_rom + '_path'] / res[br][m_rom + '_path'][-1]
                path_ref = res[br][m_ref + '_path'] / res[br][m_ref + '_path'][-1]

                # interpolate in space and time
                res_rom = interp(path_rom, res[br][f][m_rom + '_int_last'], path_ref)
                res_rom = interp(t_rom, res_rom, t_ref)

                if m_rom == '0d':
                    # add 3d results only once in loop
                    res_all[m_ref] += [res_ref]

                    # error (use 0d since generally worse than 1d)
                    err += [np.linalg.norm(res_ref - res_rom, axis=0)]

                # add rom results
                res_all[m_rom] += [res_rom]

        # error over all time steps
        err_t = np.linalg.norm(np.array(err), axis=0)

        # find time step with smallest error
        i_min = np.argmin(err_t)
        t_min = t_ref[i_min]

        # all results at time step with smallest error
        res_min = defaultdict(list)
        for m in roms:
            for br in list(res.keys()):
                res_min[m] += [res_all[m][br][:, i_min]]

        # generate "straight" centerline
        cent = straighten(read_geo(db.get_centerline_path(geo)).GetOutput(), res)

        # export for plotting
        i_3d = np.where(time['3d_rerun_last_cycle_i'])[0][i_min]
        for m in roms:
            write_rom(db, geo, cent, res_min, m)
        write_3d(db, geo, i_3d)

def write_3d(db, geo, i_3d):
    # read 3d results
    res_3d = read_geo('/home/pfaller/work/osmsc/studies/ini_1d_quad/3d_flow/' + geo + '.vtu').GetOutput()

    # read 3d mesh
    geo_3d = read_geo(db.get_volume_mesh(geo)).GetOutput()

    # get results at 3d time step
    name = f + '_' + str(db.get_3d_increment(geo) * i_3d).zfill(5)

    # get results
    array = n2v(v2n(res_3d.GetPointData().GetArray(name)) * Post().convert[f])
    array.SetName(f)

    # export
    geo_3d.GetPointData().AddArray(array)
    write_geo(out_path + '/3d/' + geo + '.vtu', geo_3d)

def write_rom(db, geo, cent, res, m_rom):
    # assign results
    branches = v2n(cent.GetPointData().GetArray('BranchId'))
    res_cent = np.zeros(cent.GetNumberOfPoints())
    for br, res_br in enumerate(res[m_rom]):
        res_cent[branches == br] = res_br * Post().convert[f]

    # set results
    array = n2v(res_cent)
    array.SetName(f)
    cent.GetPointData().AddArray(array)

    # export
    write_geo(out_path + '/' + m_rom + '/' + geo + '.vtp', cent)

def straighten(cent, res, m_rom='0d'):
    # read centerline arrays
    branches = v2n(cent.GetPointData().GetArray('BranchId'))
    area = v2n(cent.GetPointData().GetArray('CenterlineSectionArea'))
    path = v2n(cent.GetPointData().GetArray('Path'))
    points = v2n(cent.GetPoints().GetData())

    # new "linearized" arrays
    area_lin = np.zeros(cent.GetNumberOfPoints())
    points_lin = np.zeros((cent.GetNumberOfPoints(), 3))
    for br in range(np.max(v2n(cent.GetPointData().GetArray('BranchId'))) + 1):
        # all branch nodes
        i_br = branches == br
        path_br = path[i_br]
        area_br = area[i_br]
        points_br = points[i_br]

        # get branch segments
        path_rom = res[br][m_rom + '_path']

        # loop all branch segments
        for i in range(len(path_rom) - 1):
            # first and last point in segment
            p0 = np.argmin(np.abs(path_br - path_rom[i]))
            p1 = np.argmin(np.abs(path_br - path_rom[i + 1]))

            # segment points
            i_seg = np.where(i_br)[0][p0:p1+1]

            # segments for interpolation
            dist_seg = np.array([0.0, 1.0])
            path_seg = path_br[p0:p1+1].copy()
            path_seg -= path_seg[0]
            path_seg /= path_seg[-1]

            # linearly interpolate areas
            area_seg = np.array([area_br[p0], area_br[p1]])
            area_lin[i_seg] = interp(dist_seg, area_seg, path_seg)

            # linearly interpolate points
            points_seg = np.vstack((points_br[p0], points_br[p1]))
            points_lin[i_seg] = interp(dist_seg, points_seg, path_seg).T

    # add radius
    array = n2v(np.sqrt(area_lin / np.pi))
    array.SetName('segment_radius')
    cent.GetPointData().AddArray(array)

    # replace points
    points = vtk.vtkPoints()
    points.Initialize()
    for p in points_lin:
        points.InsertNextPoint(p)
    cent.SetPoints(points)

    return cent



if __name__ == '__main__':
    descr = 'Plot comparison of xd-results'
    d, g, _ = input_args(descr)
    main(d, g)



#!/usr/bin/env python

import numpy as np
import os
import sys
import pdb
import csv
from tqdm import tqdm
import seaborn as sns

from collections import defaultdict
from scipy.interpolate import CubicSpline, interp1d
from scipy.optimize import minimize

import vtk
from vtk.util.numpy_support import vtk_to_numpy as v2n
from vtk.util.numpy_support import numpy_to_vtk as n2v

from get_database import input_args, Database, Post, SimVascular
from vtk_functions import read_geo, write_geo
from get_bc_integrals import integrate_surfaces, integrate_bcs
from simulation_io import get_caps_db, collect_results, collect_results_db, get_dict
from compare_1d import add_image
from get_sv_project import coronary_sv_to_oned
from bc_0d import run_rcr, run_coronary

import matplotlib.pyplot as plt
import matplotlib.ticker as mtick
import matplotlib
# matplotlib.use('Agg')

def get_cycle(f, n_cycle):
    return np.hstack((f, np.tile(f[1:], n_cycle - 1)))


def cont_func(time, value, n_cycle):
    # repeat time for all cycles
    time_cycle = time
    for i in range(n_cycle):
        time_cycle = np.hstack((time_cycle, time[1:] + (i + 1) * time[-1]))

    # repeat value for all cycles
    value_cycle = np.hstack((value, np.tile(value[1:], n_cycle)))

    # create continues function
    fun = lambda t: interp1d(time_cycle, value_cycle)(t)

    # check if function matches origional values
    assert np.sum((fun(time) - value) ** 2) < 1e-16, 'function does not interpolate data'

    return fun, time_cycle, value_cycle


def run_0d_cycles(flow, time, p, distal_pressure, n_step=100, n_rcr=40, check=True):
    # number of cycles from rcr time constant
    if 'Rd' in p:
        t_rcr = p['C'] * p['Rd']
    elif 'R1' in p:
        t_rcr = p['C2'] * p['R2']
    else:
        raise ValueError('Unknown boundary conditions')
    n_cycle = np.max([int((t_rcr * n_rcr) // time[-1]), n_rcr])

    # 0d time
    t0 = 0.0  # best to set this as zero
    tf = time[-1] * n_cycle
    number_of_time_steps = n_step * n_cycle + 1
    time_0d = np.linspace(t0, tf, number_of_time_steps)

    # continuous inflow function
    Qfunc, _, _ = cont_func(time, flow, n_cycle)

    # output last cycle
    t_last = time[-1] * (n_cycle - 1)
    i_last = np.arange(n_step * (n_cycle - 1), n_step * n_cycle + 1, 1)
    i_prev = np.arange(n_step * (n_cycle - 2), n_step * (n_cycle - 1) + 1, 1)
    t_out = time_0d[i_last] - t_last

    # run 0d simulation
    if 'Rd' in p:
        p_0d = run_rcr(Qfunc, time_0d, p, distal_pressure)
    elif 'R1' in p:
        _, p_v_time, p_v_pres = cont_func(distal_pressure[:, 0], distal_pressure[:, 1], 3)
        p_0d = run_coronary(Qfunc, time_0d, p, p_v_time, p_v_pres, time[-1], 0.0)
    else:
        raise ValueError('Unknown boundary conditions')

    # get last cycle
    p_out = p_0d[i_last]

    # check if solution is periodic
    # delta_p = np.abs(np.mean(p_0d[i_last - 1] - p_0d[i_prev - 1]) / np.mean(p_0d[i_last - 1]))
    if check:
        delta_p = np.abs(p_out[-1] - p_out[0]) / (np.max(p_out) - np.min(p_out))
        assert delta_p < 1.0e-9, 'solution not periodic. diff=' + str(delta_p)

    return t_out, p_out


def compare_0d(db, geo, res, time, m):
    # get boundary conditions
    bc_def = db.get_bcs(geo)
    if bc_def is None:
        return None

    inlet_time = time[m]

    # get outlets
    caps = get_caps_db(db, geo)
    outlets = {}
    for cp, br in caps.items():
        if 'inflow' not in cp:
            outlets[cp] = br

    # initialize output dict
    res_bc = defaultdict(dict)

    # loop all outlets
    for j, (cp, br) in enumerate(outlets.items()):
        print('outlet ' + str(j+1) + '/' + str(len(outlets)))
        # cap bcs
        bc = bc_def['bc'][cp]
        t = bc_def['bc_type'][cp]

        # loop all cardiac cycles
        # for cycle in tqdm(range(1, time[m + '_n_cycle'] + 1)):
        for cycle in [time[m + '_n_cycle']]:
            # bc inlet flow
            inlet_flow = res[br]['flow'][m + '_all'][time[m + '_i_cycle_' + str(cycle)]]

            # output name for pressure in this cycle
            out = 'p_' + str(cycle)

            # select boundary condition
            p = {}
            if t == 'rcr':
                res_bc[br]['t'], res_bc[br][out] = run_0d_cycles(inlet_flow, inlet_time, bc, bc['Po'])
            elif t == 'resistance':
                res_bc[br]['t'] = inlet_time
                res_bc[br][out] = bc['Po'] + bc['R'] * inlet_flow
            elif t == 'coronary':
                if not bc_def['coronary']:
                    continue

                # convert coronary parameters
                cor = coronary_sv_to_oned(bc)
                p['R1'], p['R2'], p['R3'], p['C1'], p['C2'] = (cor['Ra1'], cor['Ra2'], cor['Rv1'], cor['Ca'], cor['Cc'])
                p_v_t = bc_def['coronary'][bc['Pim']][:, 0]
                p_v_p = bc_def['coronary'][bc['Pim']][:, 1]

                res_bc[br]['t'], res_bc[br][out] = run_0d_cycles(inlet_flow, inlet_time, p, np.vstack((p_v_t, p_v_p)).T)

        # copy last cycle
        res_bc[br]['p'] = res_bc[br][out]

    return res_bc


def check_bc(db, geo):
    # collect results
    # if plot_rerun and not os.path.exists(db.get_3d_flow_rerun(geo)):
    #     return

    m = '3d_rerun'
    # m = '0d'

    # get 3d results
    res, time = collect_results_db(db, geo, m)
    if m not in time:
        return

    print('Plotting ' + geo)
    # get 0d results
    if not os.path.exists(db.get_bc_0D_path(geo, m)):
        res_0d = compare_0d(db, geo, res, time, m)
        np.save(db.get_bc_0D_path(geo, m), res_0d)
    else:
        res_0d = np.load(db.get_bc_0D_path(geo, m), allow_pickle=True).item()

    inlet_time = time[m]

    # if res_0d is None or 'p_1' not in res_0d[list(res_0d.keys())[0]]:
    #     return

    # get outlets
    caps = get_caps_db(db, geo)
    outlets = {}
    for cp, br in caps.items():
        if 'inflow' not in cp:
            outlets[cp] = br

    # bounbdary condition types
    bc_def = db.get_bcs(geo)

    # get cap names
    names = db.get_cap_names(geo)

    # get numerical boundary condition time constants
    # res_num = get_dict(db.get_convergence_path())[geo]

    dpi = 300
    if len(outlets) > 50:
        dpi //= 4

    # fields = ['Flow 3D', 'Pressure 3D', 'Pressure 0D']
    # fields = ['Flow 3D', 'Pressure 3D', 'Pressure 0D', 'Pressure Paper']
    # fields = ['Pressure 0D', 'Pressure Paper']
    fields = ['Flow 3D', 'Pressure 3D']
    # fields = ['Pressure 3D', 'Pressure 0D']
    # fields = ['Pressure 3D', 'Pressure 0D', 'Pressure Q const']
    fig, ax = plt.subplots(len(fields), len(outlets), figsize=(len(outlets) * 2 + 2, 4), dpi=dpi, sharex=True, sharey='row')

    # get post-processing constants
    post = Post()

    n_max = 21
    c_max = np.min([time[m + '_n_cycle'], n_max])

    errors = []
    for j, (cp, br) in enumerate(outlets.items()):
        if br not in res_0d:
            continue
        for i, field in enumerate(fields):
            f = field.split()[0].lower()

            # plot settings
            if len(outlets) == 1:
                pos = i
            elif len(fields) == 1:
                pos = j
            else:
                pos = (i, j)
            ax[pos].grid(True)
            if j == 0:
                ax[pos].set_ylabel(field + '\n[' + post.units[f] + ']')
                ax[pos].yaxis.set_tick_params(which='both', labelleft=True)
            if i == 0:
                ax[pos].set_title(names[cp])# + ' (' + bc_def['bc_type'][cp].upper() + ')')
            if i == len(fields) - 1:
                ax[pos].set_xlabel('Time [s]')
                ax[pos].set_xlim(0, inlet_time[-1])
                ax[pos].xaxis.set_tick_params(which='both', labelbottom=True)

            # plot bcs
            for cycle in range(1, c_max + 1):
                ids = field.split()[1].lower()
                if ids == '3d':
                    x = inlet_time
                    y = res[br][f][m + '_all'][time[m + '_i_cycle_' + str(cycle)]]
                elif ids == 'q':
                    inlet_flow = res[br]['flow'][m + '_all'][time[m + '_i_cycle_' + str(cycle)]]
                    if bc_def['bc_type'][cp] == 'rcr':
                        resistance = bc_def['bc'][cp]['Rp'] + bc_def['bc'][cp]['Rd']
                    elif bc_def['bc_type'][cp] == 'coronary':
                        cor = coronary_sv_to_oned(bc_def['bc'][cp])
                        resistance = cor['Ra1'] + cor['Ra2'] + cor['Rv1']
                    x = inlet_time
                    y = np.ones(len(x)) * np.mean(inlet_flow) * resistance
                elif ids == 'paper':
                    if cycle == time[m + '_n_cycle']:
                        x = np.nan
                        y = np.nan
                    else:
                        tau = np.mean(res_num['tau'][f])
                        alpha = 1 / (np.exp(1 / tau) - 1)
                        p0 = res_0d[br]['p_' + str(cycle)]
                        p1 = res_0d[br]['p_' + str(cycle + 1)]

                        x = res_0d[br]['t']
                        y = alpha * p0 + (1 - alpha) * p1
                else:
                    raise RuntimeError('Unknown selection ' + ids)

                i_c = 1 - (cycle - 1) / (c_max - 1)
                ax[pos].plot(x, y * post.convert[f], color=plt.get_cmap('coolwarm_r')(i_c))
            if field == 'Pressure 3D':
                x = res_0d[br]['t']
                y = res_0d[br]['p']
                ax[pos].plot(x, y * post.convert[f], 'k--')

        # calculate error
        diff = interp1d(res_0d[br]['t'], res_0d[br]['p'], fill_value='extrapolate')(inlet_time) - res[br]['pressure'][m + '_cap_last']
        norm = np.max(res[br]['pressure'][m + '_cap']) - np.min(res[br]['pressure'][m + '_cap_last'])
        err = np.mean(np.abs(diff)) / norm
        errors += [err]

    max_err = np.max(errors) * 100
    max_outlet = db.get_cap_names(geo)[list(outlets.keys())[np.argmax(errors)]]

    out_str = geo + ' err=' + '{:05.2f}'.format(max_err) + '% at outlet ' + max_outlet + '\n'
    print(out_str)
    plt.gcf().suptitle(out_str)

    # save figure
    # add_image(db, geo, fig)
    f_out = db.get_bc_comparison_path(geo, m)
    fig.savefig(f_out, bbox_inches='tight')
    plt.close(fig)

    # add error to log
    db.add_bc_err(geo, m, max_err)


def check_first_last(db, geo):
    # collect results
    res, time = collect_results_db_3d_3d(db, geo)
    if res is None:
        return

    return


def plot(db, geometries):
    # read all errors
    errors = get_dict(db.get_bc_err_file('3d'))
    errors_re = get_dict(db.get_bc_err_file('3d_rerun'))

    # color by category
    colors = {'Cerebrovascular': 'k',
              'Coronary': 'r',
              'Aortofemoral': 'm',
              'Pulmonary': 'c',
              'Congenital Heart Disease': 'y',
              'Aorta': 'b',
              'Animal and Misc': '0.75'}

    err = []
    geo = []
    col = []
    for g in geometries:
        if g in errors_re:
            err += [errors_re[g]]
        elif g in errors:
            err += [errors[g]]
        else:
            continue
            # err += [np.nan]
        geo += [g]
        col += [colors[db.get_bcs(g)['params']['deliverable_category']]]

    # sort according to error
    order = np.argsort(err)
    order = np.argsort(geo)

    geo = np.array(geo)[order]
    err = np.array(err)[order]
    col = np.array(col)[order]

    geo_str = '['
    for i in np.where(err > 0.1)[0]:
        geo_str += '\'' + geo[i] + '\', '
    geo_str = geo_str[:-2] + ']'
    # print(geo_str)

    fig1, ax1 = plt.subplots(dpi=400, figsize=(15, 6))
    plt.cla()
    plt.yscale('log')
    ax1.bar(np.arange(len(err)), err, color=col)
    ax1.yaxis.grid(True)
    ax1.yaxis.set_major_formatter(mtick.PercentFormatter(xmax=100, decimals=2))
    ax1.set_ylim(0.01, 100)
    plt.xticks(np.arange(len(err)), geo, rotation='vertical')
    plt.ylabel('Max. outlet pressure error 3D vs. 0D BC')
    fname = os.path.join(db.fpath_gen, 'bc_err.png')
    # plt.legend(list(colors.keys()))
    fig1.savefig(fname, bbox_inches='tight')

    # write to csv
    with open(os.path.join(db.fpath_gen, 'bc_err.csv'), 'w', newline='') as csvfile:
        writer = csv.writer(csvfile, delimiter=',')
        for g, e in zip(geo, err):
            writer.writerow([g, str(e)])


def compare_rcr():
    params = {'Rp': 100, 'Rd': 1000, 'C': 0.001}
    distal_pressure = 0

    q_mean = 1
    nt = 1000
    nc = 10
    T = 1
    time = np.linspace(0, T*nc, nt*nc+1)
    q_var = lambda x: q_mean + np.cos(x * 2 * np.pi / T)
    q_const = lambda x: q_mean
    flows = {'var': q_var, 'const': q_const}

    fig1, ax1 = plt.subplots(dpi=400, figsize=(15, 6))

    out = defaultdict(list)
    for m, q in flows.items():
        pressure = run_rcr(q, time, params, distal_pressure)
        for i in range(nc):
            out[m] += [np.mean(pressure[np.arange(i * nt, (i + 1) * nt + 1)])]
        # pdb.set_trace()

        ax1.plot(np.arange(nc), out[m], 'o--')
    plt.show()
    pdb.set_trace()

def main(db, geometries):
    for geo in geometries:
        check_bc(db, geo)


if __name__ == '__main__':
    descr = 'Check RCR boundary condition of 3d simulation'
    d, g, _ = input_args(descr)
    main(d, g)
    # plot(d, g)
    # compare_rcr()



#!/usr/bin/env python

import pdb
import numpy as np
import os
import shutil

from get_database import Database, input_args
from vtk_functions import read_geo, write_geo


def main(db, geometries):
    # scale for geometry
    scales = {'0164_0001': 10}

    for geo in geometries:
        if geo not in scales:
            continue
        print('Fixing geometry ' + geo)

        # get points from geometry
        f_surf = db.get_surfaces(geo, 'all_exterior')
        surf = read_geo(f_surf).GetOutput()
        points = surf.GetPoints()

        # scale all points
        for i in range(points.GetNumberOfPoints()):
            point = np.array(points.GetPoint(i)) * scales[geo]
            points.SetPoint(i, tuple(point))

        # backup original file
        f_path, ext = os.path.splitext(f_surf)
        shutil.copy(f_surf, f_path + '_original' + ext)

        # save new file
        write_geo(f_surf, surf)


if __name__ == '__main__':
    descr = 'Fix wrong scale'
    d, g, _ = input_args(descr)
    main(d, g)



# Optimization for Vessel Connections

import numpy as np
from geomdl import BSpline, utilities
from geomdl.visualization import VisMPL
import numba as nb
from scipy import interpolate, optimize

@nb.jit(nopython=True)
def get_angle(V1,V2):
    return np.arccos(np.dot(-V1,V2))*(180/np.pi)

@nb.jit(nopython=True)
def get_vecs(vec):
    angles = np.zeros(vec.shape[0]-1)
    for i in range(vec.shape[0]-1):
        value = get_angle(vec[i,:],vec[i+1,:])
        angles[i] = 1/value
    return angles

@nb.jit(nopython=True)
def close_exact(data,point,radius_buffer):
    line_direction = np.zeros((data.shape[0],3))
    ss = np.zeros(data.shape[0])
    tt = np.zeros(data.shape[0])
    hh = np.zeros(data.shape[0])
    cc = np.zeros((data.shape[0],3))
    cd = np.zeros(data.shape[0])
    line_distances = np.zeros(data.shape[0])
    for i in range(data.shape[0]):
        line_direction[i,:] = (data[i,3:6] - data[i,0:3])/np.linalg.norm(data[i,3:6] - data[i,0:3])
        ss[i] = np.dot(data[i,0:3]-point,line_direction[i,:])
        tt[i] = np.dot(point-data[i,3:6],line_direction[i,:])
        d = np.array([ss[i],tt[i],0])
        hh[i] = np.max(d)
        diff = point - data[i,0:3]
        cc[i,:] = np.cross(diff,line_direction[i,:])
        cd[i] = np.linalg.norm(cc[i,:])
        line_distances[i] = np.sqrt(hh[i]**2+cd[i]**2) - data[i,6] - radius_buffer
    return line_distances

@nb.jit(nopython=True)
def get_collisions(collision_vessels,R,sample_pts,radius_buffer):
    collisions = 0
    for i in range(sample_pts.shape[0]):
        dist = close_exact(collision_vessels,sample_pts[i,:],radius_buffer)
        if np.any(dist<R):
            collisions += R - min(dist)
    return collisions

@nb.jit(nopython=True)
def check_bounds(bounds,sample_pts):
    bounds_score = 0
    for i in range(sample_pts.shape[0]):
        if sample_pts[i,0] < bounds[0,0]:
            bounds_score += (sample_pts[i,0]**2 - bounds[0,0]**2)**(1/2)
        elif sample_pts[i,0] > bounds[0,1]:
            bounds_score += (sample_pts[i,0]**2 - bounds[0,1]**2)**(1/2)
        if sample_pts[i,1] < bounds[1,0]:
            bounds_score += (sample_pts[i,1]**2 - bounds[1,0]**2)**(1/2)
        elif sample_pts[i,1] > bounds[1,1]:
            bounds_score += (sample_pts[i,1]**2 - bounds[1,1]**2)**(1/2)
        if sample_pts[i,2] < bounds[2,0]:
            bounds_score += (sample_pts[i,2]**2 - bounds[2,0]**2)**(1/2)
        elif sample_pts[i,2] > bounds[2,1]:
            bounds_score += (sample_pts[i,2]**2 - bounds[2,1]**2)**(1/2)
    return bounds_score


def connect_bezier(P1,P2,P3,P4):
    V1   = (P2 - P1)
    L1   = np.linalg.norm(V1)
    V1   = V1/np.linalg.norm(V1)
    V2   = (P4 - P3)
    L2   = np.linalg.norm(V2)
    V2   = V2/np.linalg.norm(V2)
    def create_bezier(data,V1=V1,V2=V2,P2=P2,P4=P4,L1=L1,L2=L2):
        CTR0 = P2 + (L1+data[0])*V1
        CTR1 = P4 + (L2+data[1])*V2
        CTR  = np.zeros(data.shape[0]-2+12)
        fill_CTR = np.zeros(data.shape[0]-2).reshape(-1,3)
        seg = np.linspace(0.1,0.9,num=fill_CTR.shape[0])
        for i in range(fill_CTR.shape[0]):
            fill_CTR[i,:] = CTR0*(1-seg[i]) + CTR1*(seg[i])
        fill_CTR = fill_CTR.flatten()
        CTR[0:3]     = P2
        CTR[3:6]     = CTR0
        CTR[-6:-3]   = CTR1
        CTR[-3:]     = P4
        CTR[6:-6]    += fill_CTR + data[2:]
        CTR  = CTR.reshape(-1,3).tolist()
        curve = BSpline.Curve()
        curve.degree = len(CTR) - 1
        curve.ctrlpts = CTR
        curve.knotvector = utilities.generate_knot_vector(curve.degree,len(curve.ctrlpts))
        curve.sample_size = 20
        curve.evaluate()
        return curve
    return create_bezier

def bezier_cost(data,grad,create_curve=None,R=None,
                P1=None,P3=None,collision_vessels=None,
                radius_buffer=0,bounds=None,sample_size=20):
    curve = create_curve(data)
    curve.sample_size = sample_size
    curve.evaluate()
    pts   = np.array(curve.evalpts)
    sample_pts = np.array(curve.evalpts)
    if collision_vessels is not None:
        collisions = get_collisions(collision_vessels,R,sample_pts,radius_buffer)
    else:
        collisions = 0
    bound_score = check_bounds(bounds,sample_pts)
    return bound_score+collisions

def find_optimum_connection(P1,P2,P3,P4,R,collision_vessels,bounds=None,radius_buffer=0):
    create_curve = connect_bezier(P1,P2,P3,P4)
    cost = lambda data: bezier_cost(data,None,create_curve=create_curve,R=R,P1=P1,P3=P3,collision_vessels=collision_vessels,radius_buffer=radius_buffer,bounds=bounds)
    success = False
    free_points = 1
    x0 = np.zeros(2+free_points*3)
    best = np.inf
    L = np.linalg.norm(P2-P4)
    while not success:
        lb = np.ones(x0.shape[0])*(-L)
        lb[0] = 0
        lb[1] = 0
        ub = np.ones(x0.shape[0])*L
        ctrl_bounds = []
        for b in range(len(lb)):
            ctrl_bounds.append([lb[b],ub[b]])
        res = optimize.shgo(cost,bounds=ctrl_bounds,options={'f_min':0})
        xopt = res.x
        if res.fun < 0.1:
            success = True
        else:
            free_points += 1
            x0 = np.zeros(2+free_points*3)
    curve = create_curve(xopt)
    return curve

def get_optimum_link_points(P1,P2,P3,P4,R,collision_vessels,bounds=None,radius_buffer=0):
    curve = find_optimum_connection(P1,P2,P3,P4,R,collision_vessels,bounds=bounds,radius_buffer=radius_buffer)
    pts   = np.array(curve.evalpts)
    #spline_length = np.sum(np.linalg.norm(np.diff(pts,axis=0),axis=1))
    num = 10
    curve.sample_size = num
    curve.evaluate()
    opt_pts = np.array(curve.evalpts)
    return opt_pts



#!/usr/bin/env python

import os
import vtk
import argparse
import pdb

import numpy as np

from vtk.util.numpy_support import numpy_to_vtk as n2v
from vtk.util.numpy_support import vtk_to_numpy as v2n

from .get_database import Database, input_args
from .vtk_functions import Integration, read_geo, write_geo, threshold, calculator, cut_plane
from .vtk_to_xdmf import split


def transfer_solution(node_trg, node_src, res_fields):
    """
    Transfer point data from volume mesh to surface mesh using GlobalNodeID
    Args:
        node_trg: surface mesh
        node_src: volume mesh
        res_fields: point data names to transfer
    """
    # get global node ids in both meshes
    nd_id_trg = v2n(node_trg.GetArray('GlobalNodeID')).astype(int)
    nd_id_src = v2n(node_src.GetArray('GlobalNodeID')).astype(int)

    # map volume mesh to surface mesh
    index = np.argsort(nd_id_src)
    search = np.searchsorted(nd_id_src[index], nd_id_trg)
    mask = index[search]

    # transfer results from volume mesh to surface mesh
    for i in range(node_src.GetNumberOfArrays()):
        res_name = node_src.GetArrayName(i)
        if res_name.split('_')[0] in res_fields:
            # read results from volume mesh
            res = v2n(node_src.GetArray(res_name))

            # create array to output surface mesh results
            out_array = n2v(res[mask])
            out_array.SetName(res_name)
            node_trg.AddArray(out_array)


def sort_faces(res_faces, area):
    """
    Arrange results from surface integration in matrix
    Args:
        res_faces: dictionary with key: cap id, value: result at a certain time step
        area: cross-sectional area of each cap

    Returns:
        dictionary with results as keys and matrix at all surfaces/time steps as values
    """
    # get time steps
    times = []
    for n in res_faces[list(res_faces)[0]].keys():
        times += [split(n)[0]]
    times = np.unique(np.array(times))

    # sort data in arrays according to time steps
    res_array = {'time': times}
    dim = (times.shape[0], max(list(res_faces.keys())))

    for f, f_res in res_faces.items():
        for res_name, res in f_res.items():
            time, name = split(res_name)
            if name not in res_array:
                res_array[name] = np.zeros(dim)
            res_array[name][float(time) == times, f - 1] = res

    # repeat area for all time steps to match format
    res_array['area'] = np.zeros(dim)
    for f, f_res in area.items():
        res_array['area'][:, f - 1] = f_res

    return res_array


def get_res_names(inp, res_fields):
    # result name list
    res = []

    # get integral for each result
    for i in range(inp.GetPointData().GetNumberOfArrays()):
        res_name = inp.GetPointData().GetArrayName(i)
        field = res_name.split('_')[0]
        num = res_name.split('_')[-1]

        # check if field should be added to output
        if field in res_fields:
            try:
                float(num)
                res += [res_name]
            except ValueError:
                pass

    return res


def integrate_surfaces(surf, cell_surf, res_fields, face_array='BC_FaceID'):
    """
    Integrate desired fields on all caps of surface mesh (as defined by BC_FaceID)
    Args:
        surf: reader for surface mesh
        cell_surf: surface mesh cell data
        res_fields: result fields to extract
        face_array: name of array containing face ids
    Returns:
        dictionary with result fields as keys and matrices with all faces and time steps as matrices
    """
    # generate surface normals
    normals = vtk.vtkPolyDataNormals()
    normals.SetInputData(surf)
    normals.Update()

    # recursively add calculators for normal velocities
    calc = normals
    for v in get_res_names(surf, 'velocity'):
        calc = calculator(calc, 'Normals.' + v, ['Normals', v], 'normal_' + v)

    # get all output array names
    res_names = get_res_names(surf, res_fields)

    # boundary faces
    faces = np.unique(v2n(cell_surf.GetArray(face_array)).astype(int))
    res = {}
    area = {}

    # loop boundary faces
    for f in faces:
        # skip face 0 (vessel wall)
        if f:
            # threshhold face
            thresh = threshold(calc.GetOutput(), f, face_array)

            # integrate over selected face (separately for pressure and velocity)
            integrator = Integration(thresh)

            # perform integration
            res[f] = {}
            for r in res_names:
                res[f][r] = integrator.evaluate(r)

            # store cross-sectional area
            area[f] = integrator.area()

    return sort_faces(res, area)


def integrate_bcs(fpath_surf, fpath_vol, res_fields, debug=False, debug_out='', face_array='BC_FaceID'):
    """
    Perform all steps necessary to get results averaged on caps
    Args:
        fpath_surf: surface geometry file
        fpath_vol: volume geometry file
        res_fields: results to extract
        debug: bool if debug geometry should be written
        debug_out: path for debug geometry
        face_array: name of array containing face ids
    Returns:
        dictionary with result fields as keys and matrices with all faces and time steps as matrices
    """
    if not os.path.exists(fpath_surf) or not os.path.exists(fpath_vol):
        return None

    # read surface and volume meshes
    surf = read_geo(fpath_surf).GetOutput()
    vol = read_geo(fpath_vol).GetOutput()

    # transfer solution from volume mesh to surface mesh
    transfer_solution(surf.GetPointData(), vol.GetPointData(), res_fields)

    # integrate data on boundary surfaces
    res_faces = integrate_surfaces(surf, surf.GetCellData(), res_fields, face_array=face_array)

    # write results for debugging in paraview
    if debug:
        write_geo(debug_out, surf)
    return res_faces


def main(db, geometries):
    """
    Loop all geometries in database
    """
    for geo in geometries:
        print('Processing ' + geo)

        # file paths
        fpath_surf = db.get_surfaces(geo, 'all_exterior')
        fpath_vol = db.get_volume(geo)

        bc_flow = integrate_bcs(fpath_surf, fpath_vol, ['pressure', 'velocity'])

        if bc_flow is not None:
            np.save(db.get_bc_flow_path(geo), bc_flow)


if __name__ == '__main__':
    descr = 'Plot comparison of xd-results'
    d, g, _ = input_args(descr)
    main(d, g)



#!/usr/bin/env python

import argparse
import os

import numpy as np
from collections import defaultdict

def get_dict(fpath):
    """
    Read .npy dictionary saved with numpy.save if path is defined and exists
    Args:
        fpath: path to .npy file

    Returns:
        dictionary
    """
    if fpath is not None and os.path.exists(fpath):
        return np.load(fpath, allow_pickle=True).item()
    else:
        return {}


def coronary_sv_to_oned(bc):
    """
    Convert format of coronary boundary condition parameters from svSimVascular to svOneDSolver
    """
    # unpack constants
    p1, p2, q0, q1, q2, b1 = (bc['p1'], bc['p2'], bc['q0'], bc['q1'], bc['q2'], bc['b1'])
    Rv_micro = 0.0

    # build system of equations (obtained from analytically solving for constants as defined in paper)
    # see H. J. Kim et al. "Patient-Specific Modeling of Blood Flow and Pressure in Human Coronary Arteries", p. 3198
    Ra = q2 / p2
    Ra_micro = (p1 ** 2 * q2 ** 2 - 2 * p1 * p2 * q1 * q2 + p2 ** 2 * q1 ** 2) / (
            p2 * (- q2 * p1 ** 2 + q1 * p1 * p2 - q0 * p2 ** 2 + q2 * p2))
    Ca = -p2 ** 2 / (p1 * q2 - p2 * q1)
    Cim = (- q2 * p1 ** 2 + q1 * p1 * p2 - q0 * p2 ** 2 + q2 * p2) ** 2 / ((p1 * q2 - p2 * q1) * (
            p1 ** 2 * q0 * q2 - p1 * p2 * q0 * q1 - p1 * q1 * q2 + p2 ** 2 * q0 ** 2 - 2 * p2 * q0 * q2 + p2 * q1 ** 2 + q2 ** 2))
    Rv = -(
            p1 ** 2 * q0 * q2 - p1 * p2 * q0 * q1 - p1 * q1 * q2 + p2 ** 2 * q0 ** 2 - 2 * p2 * q0 * q2 + p2 * q1 ** 2 + q2 ** 2) / (
                 - q2 * p1 ** 2 + q1 * p1 * p2 - q0 * p2 ** 2 + q2 * p2)

    # check equation residuals
    res = [p1 - (Ra_micro * Ca + (Rv + Rv_micro) * (Ca + Cim)),
           p2 - (Ca * Cim * Ra_micro * (Rv + Rv_micro)),
           q0 - (Ra + Ra_micro + Rv + Rv_micro),
           q1 - (Ra * Ca * (Ra_micro + Rv + Rv_micro) + Cim * (Ra + Ra_micro) * (Rv + Rv_micro)),
           q2 - (Ca * Cim * Ra * Ra_micro * (Rv + Rv_micro)),
           b1 - (Cim * (Rv + Rv_micro))]
    assert np.max(np.abs(res)) < 1e-5, 'SV coronary constants inconsistent'

    # export constants
    return {'Ra1': Ra, 'Ra2': Ra_micro, 'Ca': Ca, 'Cc': Cim, 'Rv1': Rv, 'P_v': 0.0}


def rec_dict():
    """
    Recursive defaultdict
    """
    return defaultdict(rec_dict)



# maintainer: Martin Pfaller
# 
# created by Martina Weigl
# adapted from Matlab toolbox
# http://de.mathworks.com/matlabcentral/fileexchange/27826-fast-assembly-of-stiffness-and-matrices-in-finite-element-method-using-nodal-elements
# http://www.mis.mpg.de/preprints/tr/report-1111.pdf

import sys, timeit
from numpy import abs, array, ones, ravel, squeeze, sum, zeros
import numpy as np
from scipy.special import factorial
from scipy import sparse
from scipy.sparse.linalg import spsolve, splu

# help functions for stiffness matrix calculation
def phider(coord,point,etype,nargout):
    
    jacout = False
    if nargout > 2:
        jacout = True
    detout = False
    if nargout > 1:
        detout = True
        
    nod = coord.shape[1]
    nop = point.shape[1]
    nos = coord.shape[2]
    noe = coord.shape[0]
     
    # gradients of the shape functions
    dshape = shapeder(point,etype)  
    
    if jacout == True:
        jac = zeros((noe,nop,nod,nod))
    if detout == True:
        detj = zeros((noe,nop,1))
    dphi = zeros((noe,nop,nod,nos))

    for poi in range(nop):
         
        tjac = smamt(dshape[poi],coord)
        tjacinv, tjacdet = aminv(tjac)
        dphi[:,poi,:,:] = amsm(tjacinv,dshape[poi])
        
        if jacout == True:
            jac[:,poi,:,:] = tjac
        if detout == True:
            detj[:,0,:] = abs(tjacdet)
        
    return dphi, detj

    
def shapeder(point,etype):
     
    nod = point.shape[0]
    nop = point.shape[1]

    if nod == 1:
         
        l1 = point[0]
        l2 = 1 - l1
         
        if etype == "P0":
              
            dshape = zeros((nop,1,1))
             
        elif etype == "P1":
            
            dshape = array([1, -1])
            dshape = dshape.reshape(2,1)
            dshape = dshape * ones((nop,1))
            dshape = dshape.reshape(nop,1,2)
            
        elif etype == "P2":
            
            dshape = array([[4 * l1 - 1],
                            [-4 * l2 + 1],
                            [4 * (l2 - l1)]])
            dshape = dshape.reshape(nop,1,3)
            
        else:
            print("Error: Only P1, P2 elements implemented.")
    
    
    if nod == 2:
        
        l1 = point[0]
        l2 = point[1]
        l3 = 1 - l1 - l2
        
        if etype == "P0":
            
            dshape = zeros((nop,2,1))
            
        elif etype == "P1":
            
            dshape = array([[1, 0, -1],
                            [0, 1, -1]])
            dshape = dshape.reshape(6,1)
            dshape = dshape * ones((nop,1))
            dshape = dshape.reshape(nop,2,3)
            
        elif etype == "P2":
            
            dshape = array([[- 4 * l3 + 1],
                            [- 4 * l3 + 1],
                            [4 * l1 - 1],
                            [zeros((nop,1))],
                            [zeros((nop,1))],
                            [4 * l2 - 1],
                            [4 * l2],
                            [4 * l1],
                            [-4 * l2],
                            [4 * (l3 - l2)],
                            [4 * (l3 - l1)],
                            [-4 * l1]])
            dshape = dshape.reshape(nop,2,6)
            
        else:
            print("Error: Only P1, P2 elements implemented.")
            
            
    if nod == 3:
        
        l1 = point[0]
        l2 = point[1]
        l3 = point[2]
        l4 = 1 - l1 - l2 - l3
        
        if etype == "P0":
            
            dshape = zeros((nop,1,1))
            
        elif etype == "P1":
            
            dshape = array([[1, 0, 0, -1],
                            [0, 1, 0, -1],
                            [0, 0, 1, -1]])
            dshape = dshape.reshape(12,1)
            dshape = dshape * ones((nop,1))
            dshape = dshape.reshape(nop,3,4)
            
        elif etype =="P2":
            
            dshape = array([[-4 * l4 + 1],
                            [-4 * l4 + 1],
                            [-4 * l4 + 1],
                            [4 * l1 - 1],
                            [zeros((nop,1))],
                            [zeros((nop,1))],
                            [zeros((nop,1))],
                            [4 * l2 - 1],
                            [zeros((nop,1))],
                            [zeros((nop,1))],
                            [zeros((nop,1))],
                            [4 * l3 - 1],
                            [4 * (l4 - l1)],
                            [-4 * l1],
                            [-4 * l1],
                            [-4 * l2],
                            [4 * (l4 - l2)],
                            [-4 * l2],
                            [-4 * l3],
                            [-4 * l3],
                            [4 * (l4 - l3)],
                            [4 * l2],
                            [4 * l1],
                            [zeros((nop,1))],
                            [4 * l3],
                            [zeros((nop,1))],
                            [4 * l1],
                            [zeros((nop,1))],
                            [4 * l3],
                            [4 * l2]])
            dshape = dshape.reshape(nop,3,10)
            
        elif etype == "Q1":
             
            x = point[0]
            y = point[1]
            z = point[2]
             
            S = array([[0,0,0],
                       [1,0,0],
                       [1,1,0],
                       [0,1,0],
                       [0,0,1],
                       [1,0,1],
                       [1,1,1],
                       [0,1,1]]) * 2 - 1
                        
            dshape = zeros((nop,3,8))
             
            for n in range(nop):
                for m in range(8):
                    dshape[n][0][m] = 1./8 * S[m][0] * (1 + S[m][1] * y) * (1 + S[m][2] * z)
                    dshape[n][1][m] = 1./8 * (1 + S[m][0] * x) * S[m][1] * (1 + S[m][2] * z)
                    dshape[n][2][m] = 1./8 * (1 + S[m][0] * x) * (1 + S[m][1] * y) * S[m][2]
                    
        else:
            print("Error: Only P1, P2, Q1 elements implemented.")
     
                
    return dshape        


def smamt(smx,ama):
    
    ny = ama.shape[1]
    nx = ama.shape[2]
    nz = ama.shape[0]
    
    nk = smx.shape[0]
    
    amb = zeros((nz,nk,ny))
    for row in range(nk):
        
        amb[:,row,:] = svamt(smx[row],ama)
        
    return amb
    
    
def svamt(svx,ama):
    
    ny = ama.shape[1]
    nx = ama.shape[2]
    nz = ama.shape[0]
    
    avx = zeros((nz,ny,nx))
    
    for n in range(nz):
        avx[n] = svx
        
    avb = ama * avx
    avb = sum(avb, axis=2)    
   
    return avb


def aminv(ama):
    
    nx = ama.shape[1]
    nz = ama.shape[0]
    
    
    if nx == 1:
        
        dem = squeeze(ama)
        
        amb = zeros((nz,nx,nx))
        for n in range(nz):
            amb[n] = 1./ama[n]
  
        return amb, dem
    
    
    elif nx == 2:
        
        x1,x2,y1,y2 = zeros((nz,1)),zeros((nz,1)),zeros((nz,1)),zeros((nz,1))
        
        for n in range(nz):
            x1[n],x2[n] = squeeze(ama[n][0][0]), squeeze(ama[n][0][1])
            y1[n],y2[n] = squeeze(ama[n][1][0]), squeeze(ama[n][1][1])
            
        dem = x1 * y2 - x2 * y1
            
        amb = zeros((nz,nx,nx))
        for n in range(nz):
            amb[n][0][0] = y2[n] / dem[n]
            amb[n][1][1] = x1[n] / dem[n]
            amb[n][0][1] = -x2[n] / dem[n]
            amb[n][1][0] = -y1[n] / dem[n]
            
        return amb, dem
    
    
    elif nx == 3:
        
        x1,x2,x3,y1,y2,y3,z1,z2,z3 = zeros((nz,1)),zeros((nz,1)),zeros((nz,1)),zeros((nz,1)),zeros((nz,1)),zeros((nz,1)),zeros((nz,1)),zeros((nz,1)),zeros((nz,1))
        
        for n in range(nz):
            x1[n],x2[n],x3[n] = squeeze(ama[n][0][0]), squeeze(ama[n][0][1]), squeeze(ama[n][0][2])
            y1[n],y2[n],y3[n] = squeeze(ama[n][1][0]), squeeze(ama[n][1][1]), squeeze(ama[n][1][2])
            z1[n],z2[n],z3[n] = squeeze(ama[n][2][0]), squeeze(ama[n][2][1]), squeeze(ama[n][2][2])  
            
        dem = x1 * (y2 * z3 - y3 * z2) - x2 * (y1 * z3 - y3 * z1) + x3 * (y1 * z2 - y2 * z1)  
        
        C11 = y2 * z3 - y3 * z2
        C12 = -(y1 * z3 - y3 * z1)
        C13 = y1 * z2 - y2 * z1
        C21 = -(x2 * z3 - x3 * z2)
        C22 = x1 * z3 - x3 * z1
        C23 = -(x1 * z2 - x2 * z1)
        C31 = x2 * y3 - x3 * y2
        C32 = -(x1 * y3 - x3 * y1)
        C33 = x1 * y2 - x2 * y1
        
        amb = zeros((nz,nx,nx))
        for n in range(nz):
            amb[n][0][0] = C11[n] / dem[n]
            amb[n][1][0] = C12[n] / dem[n]
            amb[n][2][0] = C13[n] / dem[n]
            amb[n][0][1] = C21[n] / dem[n]
            amb[n][1][1] = C22[n] / dem[n]
            amb[n][2][1] = C23[n] / dem[n]
            amb[n][0][2] = C31[n] / dem[n]
            amb[n][1][2] = C32[n] / dem[n]
            amb[n][2][2] = C33[n] / dem[n]
            
        return amb, dem
    
    else:
        print("Error: Array operation for inverting matrices of dimension more than three is not available; performance may be bad.")
        sys.exit()


def amsm(ama,smx):  
    
    nx = ama.shape[1]
    ny = ama.shape[2]
    nz = ama.shape[0]
    
    nk = smx.shape[1]
    
    amb = zeros((nz,nx,nk))
    for col in range(nk):
        amb[:,:,col] = amsv(ama,smx[:,col])
        
    return amb
              
            
def amsv(ama,svx):
    
    nx = ama.shape[1]
    ny = ama.shape[2]
    nz = ama.shape[0]
            
    avx = zeros((nz,nx,ny))
    
    for n in range(nz):
        avx[n] = svx
        
    avb = ama * avx
    avb = sum(avb, axis=2) 

    return avb


def astam(asx, ama):
    
    nx = ama.shape[1]
    ny = ama.shape[2]
    nz = ama.shape[0]
    
    asm = zeros((nz,nx,ny))
    
    for n in range(nz):
        asm[n] = asx[n]
    
    amb = ama * asm
    
    return amb


def amtam(amx,ama):
    
    ny = ama.shape[2]
    
    nk = amx.shape[2]
    nz = amx.shape[0]
 
    amb = zeros((nz,nk,ny))
    
    for row in range(nk):
        amb[:,row] = avtam(amx[:,:,row],ama)
      
    return amb


def avtam(avx,ama):
    
    nx = ama.shape[1]
    ny = ama.shape[2]
    nz = ama.shape[0]

    avm = zeros((nz,nx,ny))
    
    for n in range(nz):
        for m in range(ny):
            avm[n,:,m] = avx[n]
           
    avb = ama * avm
    avb = sum(avb, axis=1)

    return avb


# class to store StiffnessMatrix and solve laplace problem
class StiffnessMatrix:
    # create stiffness matrix on initialization and store as class variable
    def __init__(self, element, coordinates, DOF_d = []):
        start = timeit.default_timer()
        
        # numder of nodes, problem dimension
        NN, DIM = coordinates.shape
        
        # number of elements, number of local basic functions
        NE, NLB = element.shape
        
        coord = zeros((NE,DIM,NLB))
        for x in range(NE):
            for n in range(NLB):
                for m in range(DIM):
                    coord[x][m][n] = coordinates[element[x][n]][m]
        
        # the coordinates of the points on the reference element
        IP = array([[0.25],
                    [0.25],
                    [0.25]])
    
        # element type for shape function       
        etype = "P1"
    
        # gradients of the basis functions
        nargout = 2
        dphi, detj = phider(coord,IP,etype,nargout)
        areas = abs(squeeze(detj)) / factorial(DIM, exact = True)
        
        dphi = squeeze(dphi)
        
        Z = astam(areas,amtam(dphi,dphi))
        
        Y = zeros((NE,NLB,NLB))
        for n in range(NE):
            for m in range(NLB):
                Y[n][m] = element[n][m]
                
        X = zeros((NE,NLB,NLB))
        for n in range(NE):
            X[n] = element[n]
        
        # stiffness matrix
        A = sparse.csc_matrix((ravel(Z),(ravel(Y),ravel(X))),shape=[NN,NN])

        # store number of nodes
        self._n = A.shape[0]
        
        # store stiffness matrix
        self._A = A
        
        stop = timeit.default_timer()
        print('created stiffness matrix in %.1fs' %(stop-start))
        
        # check if StiffnessMatrix is initialized with dirichlet boundary conditions
        if any(DOF_d):
            start = timeit.default_timer()
            
            # detect free DOFs
            DOF_f = np.setdiff1d(range(self._n),DOF_d)
            
            # LU decomposition for sparse matrix
            self._LU = splu(A[DOF_f,:][:,DOF_f])
            
            # store free DOFs with which LU-decomposition was initialized
            self._DOF_f_init = DOF_f
        
            stop = timeit.default_timer()
            print('created LU-decomposition in %.1fs' %(stop-start))
    
    # harmonic lifting through solving the laplace equation
    def HarmonicLift(self, DOF_d, X_d):
        # detect free DOFs
        DOF_f = np.setdiff1d(range(self._n),DOF_d)

        # condensate dirichlet DOFs
        rhs = -self._A[DOF_f,:][:,DOF_d].dot(X_d)
        
        # solve harmonic lifting
        if hasattr(self, '_LU') and np.array_equal(np.sort(DOF_f), np.sort(self._DOF_f_init)):
            # DBC DOFs are equal to the ones with which StiffnessMatrix was initialized
            # -> use LU-decomposition to solve (fast)
            X_f = self._LU.solve(rhs)
        else:
            # different DBC DOFs
            # -> solve by inverting A (slow)
            X_f = spsolve(self._A[DOF_f,:][:,DOF_f],rhs)
        
        # assemble solution vector
        X = zeros((self._n))
        X[DOF_d] = X_d
        X[DOF_f] = X_f

        return X


#!/usr/bin/env python

import numpy as np
import os
import pdb

from scipy.interpolate import CubicSpline, interp1d
from scipy.optimize import minimize

import vtk
from vtk.util.numpy_support import vtk_to_numpy as v2n
from vtk.util.numpy_support import numpy_to_vtk as n2v

from get_database import input_args, Database, Post, SimVascular
from vtk_functions import read_geo, write_geo
from get_bc_integrals import integrate_surfaces, integrate_bcs

import matplotlib.pyplot as plt


def fourier(x, n_sample_freq=128):
    """
    Inverse fourier transformation from frequencies in x (real, imaginary)
    """
    assert x.shape[0] % 2 == 0, 'odd number of parameters'
    n_mode = x.shape[0] // 2

    x_complex = x[:n_mode] + 1j * x[n_mode:]
    inflow_fft = np.zeros(n_sample_freq + 1, dtype=complex)
    inflow_fft[:n_mode] = x_complex

    return np.fft.irfft(inflow_fft)


def error(time, inflow, time_smooth, inflow_smooth):
    """
    Get error between input inflow and smooth inflow
    """
    # repeat last value at the start
    time_smooth = np.insert(time_smooth, 0, 0)
    inflow_smooth = np.insert(inflow_smooth, 0, inflow_smooth[-1])

    # interpolate to coarse time
    inflow_interp = interp1d(time_smooth, inflow_smooth)(time)

    return np.sqrt(np.sum((inflow - inflow_interp) ** 2))


def optimize_inflow(time, inflow, n_mode=10, n_sample_real=256, debug=False):
    """
    Optimize fourier-smoothed inflow to interpolate input inflow
    """
    # define fourier smoothing
    assert n_sample_real % 2 == 0, 'odd number of samples'
    n_sample_freq = n_sample_real // 2

    # insert last 3d time step as 1d initial condition (periodic solution)
    time = np.insert(time, 0, 0)
    inflow = np.insert(inflow, 0, inflow[-1])

    # linearly interpolate at fine time points
    time_smooth = np.linspace(0, time[-1], n_sample_real + 1)[1:]
    inflow_interp_lin = interp1d(time, inflow)(time_smooth)

    # get starting value from fft
    inflow_fft = np.fft.rfft(inflow_interp_lin)
    x0 = inflow_fft[:n_mode]
    x0_split = np.array(np.hstack((np.real(x0), np.imag(x0))))

    # setup otimization problem
    run = lambda x: error(time, inflow, time_smooth, fourier(x, n_sample_freq))

    # optimize frequencies to match inflow profile
    res = minimize(run, x0_split, tol=1.0e-8, options={'disp': debug})
    inflow_smooth = fourier(res.x, n_sample_freq)

    # add time step zero
    time_smooth = np.insert(time_smooth, 0, 0)
    inflow_smooth = np.insert(inflow_smooth, 0, inflow_smooth[-1])

    # re-sample to n_sample_real
    time_out = np.linspace(0, time_smooth[-1], n_sample_real)
    inflow_out = interp1d(time_smooth, inflow_smooth)(time_out)

    if debug:
        import matplotlib.pyplot as plt
        fig, ax = plt.subplots(dpi=300, figsize=(12, 6))
        ax.plot(time_smooth[1:], fourier(x0_split), 'b-')
        ax.plot(time_smooth, inflow_smooth, 'r-')
        ax.plot(time, inflow, 'kx')
        plt.grid()
        plt.show()

    return time_out, inflow_out


def error_phase(t1, f1, t2, f2, dt):
    # period
    p = t2[-1]

    # shift and replicate
    t2_cyc = np.hstack((t2[:-1] - p, t2, t2[1:] + p)) + dt
    f2_cyc = np.hstack((f2[:-1], f2, f2[1:]))

    # interpolate to coarse time
    f2_interp = interp1d(t2_cyc, f2_cyc)(t1)

    return np.sum((f1 - f2_interp) ** 2)


def phase_shift(t, f, dt):
    # period
    p = t[-1]

    # shift and replicate
    t_cyc = np.hstack((t[:-1] - p, t, t[1:] + p)) + dt
    f_cyc = np.hstack((f[:-1], f, f[1:]))

    # interpolate to original time
    return interp1d(t_cyc, f_cyc)(t)


def optimize_phase(t1, f1, t2, f2, debug=False):
    # setup otimization problem
    run = lambda x: error_phase(t1, f1, t2, f2, x)

    # find time shift to match inflow profile
    res = minimize(run, [0.15], tol=1.0e-8, options={'disp': debug}, bounds=((-t2[-1]/4, t2[-1]/4), ))

    return res.x[0]


def read_velocity(f_dat):
    """
    Read velocity, time steps, and node ids from bct.dat
    """
    # read text file
    with open(f_dat) as f:
        lines = f.readlines()

    # get number of points and time steps
    n_p, n_t = (int(l) for l in lines[0].strip().split())

    # read points
    vel = []
    time = []
    points = []
    coords = []
    for i in range(n_p):
        # line of point header
        split = lines[1 + i * (n_t + 1)].strip().split()

        # point coordinates
        coords += [[float(split[i]) for i in range(3)]]

        # point id
        points += [int(split[-1])]

        # read time steps
        vel_p = []
        for j in range(n_t):
            # line of time step
            split = lines[2 + i + i * n_t + j].split()

            # velocity vector
            vel_p += [[float(split[i]) for i in range(3)]]

            # time
            if i == 0:
                time += [float(split[-1])]

        vel += [vel_p]
    return np.array(vel), np.array(time), np.array(points), np.array(coords)


def write_velocity(f_dat, vel, time, points, coords):
    """
    Write bct.dat file
    """
    # get dimensions
    n_p, n_t, dim = vel.shape

    assert n_p == points.shape[0], 'number of points mismatch'
    assert n_p == coords.shape[0], 'number of coordinates mismatch'
    assert n_t == time.shape[0], 'number of time steps mismatch'
    assert dim == 3, 'number of dimensions mismatch'
    assert coords.shape[1] == 3, 'number of dimensions mismatch'

    with open(f_dat, 'w+') as f:
        # write header
        f.write(str(n_p) + ' ' + str(n_t) + '\n')

        # write points
        for i in range(n_p):
            # write point
            for j in range(3):
                f.write("{:.6e}".format(coords[i, j]) + ' ')
            f.write(str(n_t) + ' ' + str(points[i]) + '\n')

            # write time steps
            np.savetxt(f, np.vstack((vel[i].T, time)).T, fmt='%1.6e')


def add_velocity(inlet, vel, time, points):
    """
    Add velocity vectors to inlet geometry bct.vtp
    """
    # get unique point ids
    ids = v2n(inlet.GetPointData().GetArray('GlobalNodeID'))

    # remove all point arrays except GLobalNodeId
    names = [inlet.GetPointData().GetArrayName(i) for i in range(inlet.GetPointData().GetNumberOfArrays())]
    for n in names:
        if n != 'GlobalNodeID':
            inlet.GetPointData().RemoveArray(n)

    # add velocity vectors to nodes
    for i, t in enumerate(time):
        # create new array for time step
        array = vtk.vtkDoubleArray()
        array.SetNumberOfComponents(3)
        array.SetNumberOfTuples(vel.shape[0] * 3)
        array.SetName('velocity_' + str(t))
        inlet.GetPointData().AddArray(array)

        # fill array
        for j, p in enumerate(points):
            k = np.where(ids == p)[0][0]
            v = vel[j, i]
            array.SetTuple3(k, v[0], v[1], v[2])


def integrate_inlet(f_in):
    """
    Get inlet flow from bct.dat and bct.vtp
    """
    # read inlet geometry from bct.vtp
    inlet = read_geo(f_in + '.vtp').GetOutput()

    # integrate over inlet
    return integrate_surfaces(inlet, inlet.GetCellData(), 'velocity', face_array='ModelFaceID')


def overwrite_inflow(db, geo, n_sample_real=256):
    """
    Overwrite bct.dat and bct.vtp from svpre with own high-fidelity inflow
    """
    # define project paths
    f_in = os.path.join(db.get_solve_dir_3d(geo), 'bct')

    # read inflow from file
    time, inflow = db.get_inflow_osmsc(geo)

    # fit inflow using fourier smoothing
    time, inflow = optimize_inflow(time, inflow, n_sample_real)

    # read constant inflow
    vel_dat, time_dat, points, coords = read_velocity(f_in + '.dat')

    # integrate inflow from from bct.dat and bct.vtp
    surf_int = integrate_inlet(f_in)

    # scale velocity
    pdb.set_trace()
    vel_scaled = vel_dat / surf_int['velocity'] * np.expand_dims(inflow, axis=1)

    # overwrite bct.dat
    write_velocity(f_in + '.dat', vel_scaled, time, points, coords)

    # overwrite bct.vtp
    inlet = read_geo(f_in + '.vtp').GetOutput()
    add_velocity(inlet, vel_scaled, time, points)
    write_geo(f_in + '.vtp', inlet)


def check_inflow(db, geo):
    post = Post()

    # create output folder
    check_dir = os.path.join(db.get_solve_dir_3d(geo), 'check')
    os.makedirs(os.path.join(check_dir), exist_ok=True)

    # define project paths
    f_in = os.path.join(db.get_solve_dir_3d(geo), 'bct')
    f_out_fig = os.path.join(check_dir, geo + '_inflow')
    f_out_vtp = os.path.join(check_dir, 'initial.vtp')

    # read inflow from file
    time, inflow = db.get_inflow(geo)
    time_smooth, inflow_smooth = db.get_inflow_smooth(geo)

    # get model inlet from bct.dat and bct.vtp
    surf_int = integrate_inlet(f_in)

    # postproc initial conditions
    sv = SimVascular()
    sv.run_post(db.get_solve_dir_3d(geo), ['-start', '0', '-stop', '0', '-incr', '1', '-vtkcombo', '-vtp', 'check/initial.vtp'])

    # get initial conditions
    fpath_surf = os.path.join(db.get_solve_dir_3d(geo), 'mesh-complete', 'mesh-surfaces', 'inflow.vtp')
    ini = integrate_bcs(fpath_surf, f_out_vtp, ['pressure', 'velocity'], face_array='ModelFaceID')

    # plot comparison
    fig, ax = plt.subplots(dpi=300, figsize=(12, 6))
    plt.plot(time_smooth, inflow_smooth * post.convert['flow'], 'g-')
    plt.plot(surf_int['time'], surf_int['velocity'][:, -1] * post.convert['flow'], 'r--')
    plt.plot(0, ini['velocity'][0][-1] * post.convert['flow'], 'bo', fillstyle='none')
    plt.plot(time, inflow * post.convert['flow'], 'kx')
    plt.xlabel('Time [s]')
    plt.ylabel('Flow [l/h]')
    plt.title('Initial pressure ' + '{:2.1f}'.format(ini['pressure'][0][-1] * post.convert['pressure']) + ' mmHg')
    plt.grid()
    ax.legend(['Optimized for rerun', 'SimVascular', 'Initial condition for rerun', 'OSMSC'])
    fig.savefig(f_out_fig, bbox_inches='tight')
    plt.cla()


def fix_inflows(db, geo):
    plot = False

    t1, f1 = db.get_inflow(geo)
    if t1 is None:
        return

    if geo == '0174_0000':
        geo_in = '0176_0000'
    elif geo == '0176_0000':
        geo_in = '0174_0000'
    else:
        geo_in = geo
    t2, f2 = db.get_inflow_osmsc(geo_in)

    if len(t2) - 1 == len(t1):
        # optimize an inflow profile
        print('  only coarse inflow available')
        t2, f2 = optimize_inflow(t1, f1)
        plot = True
    else:
        # minimize difference between inflows by shifting in time
        dt = optimize_phase(t1, f1, t2, f2)
        f2 = phase_shift(t2, f2, dt)

    # interpolate to coarse time steps
    f12 = interp1d(t2, f2)(t1)

    # fix scaling
    if geo == '0174_0000' or geo == '0176_0000':
        a = np.min(f1) / np.min(f12)
        f12 *= a
        f2 *= a

    # error between flow profiles
    err = np.max(np.abs(f1 - f12)) / np.abs(np.mean(f1))

    print(geo, "{:.2e}".format(err))

    # save
    if err < 0.1:
        np.savetxt(db.get_inflow_smooth_path(geo), np.vstack((t2, f2)).T)

    if plot:
        fig, ax = plt.subplots(dpi=300, figsize=(12, 6))
        post = Post()
        plt.plot(t1, f1 * post.convert['flow'], 'k-')
        plt.plot(t2, f2 * post.convert['flow'], 'r--')
        plt.xlabel('Time [s]')
        plt.ylabel('Flow [l/h]')
        ax.legend(['Inflow from results', 'Inflow from extras'])
        ax.grid(True)
        f_out = os.path.join('/home/pfaller/work/osmsc/data_generated/check_inflows', geo)
        fig.savefig(f_out, bbox_inches='tight')
        plt.close(fig)

    return False


def main(db, geometries):
    for geo in geometries:
        # print('Checking geometry ' + geo)
        # check_inflow(db, geo)
        fix_inflows(db, geo)

if __name__ == '__main__':
    descr = 'Check inlet flow of 3d simulation'
    d, g, _ = input_args(descr)
    main(d, g)



# smooth connections between trees in a network
from geomdl import BSpline
from geomdl import utilities
from geomdl.visualization import VisMPL
import numpy as np
from copy import deepcopy
#from .optimize_connection_v3 import *
#from .optimize_connection_v4 import *
from .optimize_connection_v5 import *
from tqdm import trange
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D

def get_initial_vectors(forest,network,assignment,tree,edge):
    # Terminal Vessel Vector
    P1 = forest.networks[network][tree].data[assignment[edge],0:3]
    P2 = forest.networks[network][tree].data[assignment[edge],3:6]
    V1 = P2 - P1
    V1 = V1/np.linalg.norm(V1)

    # Connection Vessel Vector
    P3 = forest.networks[network][tree].data[assignment[edge],3:6]
    P4 = forest.connections[network][0][edge]
    V2 = P4 - P3
    V2 = V2/np.linalg.norm(V2)
    return V1,V2,P4

def link_v3(forest,network_id,tree_1,tree_2,radius_buffer):
    tmp_tree_connections = tree_connections(forest,network_id,tree_1,tree_2,radius_buffer)
    tmp_tree_connections.solve()
    tmp_tree_connections.build_vessels()
    count_1 = 0
    count_2 = 0
    tree_1_connections = []
    tree_2_connections = []
    for i in range(len(tmp_tree_connections.connection_solutions)):
        forest.networks[network_id][tree_1].data[tmp_tree_connections.connection_solutions[i].vessel_1_id,0:3] = tmp_tree_connections.connection_solutions[i].vessels[0][0,0:3]
        forest.networks[network_id][tree_1].data[tmp_tree_connections.connection_solutions[i].vessel_1_id,3:6] = tmp_tree_connections.connection_solutions[i].vessels[0][0,3:6]
        forest.networks[network_id][tree_1].data[tmp_tree_connections.connection_solutions[i].vessel_1_id,12:15] = tmp_tree_connections.connection_solutions[i].vessels[0][0,12:15]
        forest.networks[network_id][tree_1].data[tmp_tree_connections.connection_solutions[i].vessel_1_id,20] = tmp_tree_connections.connection_solutions[i].vessels[0][0,20]
        tmp_tree_connections.connection_solutions[i].vessels[0][0,21] = forest.networks[network_id][tree_1].data[tmp_tree_connections.connection_solutions[i].vessel_1_id,21]
        tmp_tree_connections.connection_solutions[i].vessels[0][0,22] = forest.networks[network_id][tree_1].data[tmp_tree_connections.connection_solutions[i].vessel_1_id,22]
        tmp_tree_connections.connection_solutions[i].vessels[0][0,26] = forest.networks[network_id][tree_1].data[tmp_tree_connections.connection_solutions[i].vessel_1_id,26]
        tmp_tree_connections.connection_solutions[i].vessels[0][0,28] = forest.networks[network_id][tree_1].data[tmp_tree_connections.connection_solutions[i].vessel_1_id,28]
        parent_vessel_index = int(tmp_tree_connections.connection_solutions[i].vessel_1_id)
        next_vessel_index_1   = int(max(forest.networks[network_id][tree_1].data[:,-1]))+1+count_1
        previous_node_index = int(forest.networks[network_id][tree_1].data[tmp_tree_connections.connection_solutions[i].vessel_1_id,19])
        next_node_index     = int(max(forest.networks[network_id][tree_1].data[:,19]))+1
        forest.networks[network_id][tree_1].data[tmp_tree_connections.connection_solutions[i].vessel_1_id, 15] = next_vessel_index_1 #added
        forest.networks[network_id][tree_1].data[tmp_tree_connections.connection_solutions[i].vessel_1_id, 23] = 1 # added
        forest.networks[network_id][tree_1].data[tmp_tree_connections.connection_solutions[i].vessel_1_id, 24] = 0 # added
        for vdx in range(1,tmp_tree_connections.connection_solutions[i].vessels[0].shape[0]):
            tmp_tree_connections.connection_solutions[i].vessels[0][vdx,-1] = next_vessel_index_1
            tmp_tree_connections.connection_solutions[i].vessels[0][vdx,17] = parent_vessel_index
            tmp_tree_connections.connection_solutions[i].vessels[0][vdx,18] = previous_node_index
            tmp_tree_connections.connection_solutions[i].vessels[0][vdx,19] = next_node_index
            tmp_tree_connections.connection_solutions[i].vessels[0][vdx,21] = tmp_tree_connections.connection_solutions[i].vessels[0][vdx-1,21]
            tmp_tree_connections.connection_solutions[i].vessels[0][vdx,22] = tmp_tree_connections.connection_solutions[i].vessels[0][vdx-1,22]
            tmp_tree_connections.connection_solutions[i].vessels[0][vdx,26] = tmp_tree_connections.connection_solutions[i].vessels[0][vdx-1,26]+1
            tmp_tree_connections.connection_solutions[i].vessels[0][vdx,28] = tmp_tree_connections.connection_solutions[i].vessels[0][vdx-1,28]
            tmp_tree_connections.connection_solutions[i].vessels[0][vdx-1,15] = next_vessel_index_1
            tmp_tree_connections.connection_solutions[i].vessels[0][vdx-1,23] = 1
            tmp_tree_connections.connection_solutions[i].vessels[0][vdx-1,24] = 0
            parent_vessel_index = next_vessel_index_1
            next_vessel_index_1 += 1
            previous_node_index = next_node_index
            next_node_index += 1
            count_1 += 1
        forest.networks[network_id][tree_2].data[tmp_tree_connections.connection_solutions[i].vessel_2_id,0:3] = tmp_tree_connections.connection_solutions[i].vessels[1][0,0:3]
        forest.networks[network_id][tree_2].data[tmp_tree_connections.connection_solutions[i].vessel_2_id,3:6] = tmp_tree_connections.connection_solutions[i].vessels[1][0,3:6]
        forest.networks[network_id][tree_2].data[tmp_tree_connections.connection_solutions[i].vessel_2_id,12:15] = tmp_tree_connections.connection_solutions[i].vessels[1][0,12:15]
        forest.networks[network_id][tree_2].data[tmp_tree_connections.connection_solutions[i].vessel_2_id,20] = tmp_tree_connections.connection_solutions[i].vessels[1][0,20]
        tmp_tree_connections.connection_solutions[i].vessels[1][0,21] = forest.networks[network_id][tree_2].data[tmp_tree_connections.connection_solutions[i].vessel_2_id,21]
        tmp_tree_connections.connection_solutions[i].vessels[1][0,22] = forest.networks[network_id][tree_2].data[tmp_tree_connections.connection_solutions[i].vessel_2_id,22]
        tmp_tree_connections.connection_solutions[i].vessels[1][0,26] = forest.networks[network_id][tree_2].data[tmp_tree_connections.connection_solutions[i].vessel_2_id,26]
        tmp_tree_connections.connection_solutions[i].vessels[1][0,28] = forest.networks[network_id][tree_2].data[tmp_tree_connections.connection_solutions[i].vessel_2_id,28]
        parent_vessel_index = int(tmp_tree_connections.connection_solutions[i].vessel_2_id)
        next_vessel_index_2   = int(max(forest.networks[network_id][tree_2].data[:,-1]))+1+count_2
        previous_node_index = int(forest.networks[network_id][tree_2].data[tmp_tree_connections.connection_solutions[i].vessel_2_id,19])
        next_node_index     = int(max(forest.networks[network_id][tree_2].data[:,19]))+1
        forest.networks[network_id][tree_2].data[tmp_tree_connections.connection_solutions[i].vessel_2_id, 15] = next_vessel_index_2 #added
        forest.networks[network_id][tree_2].data[tmp_tree_connections.connection_solutions[i].vessel_2_id, 23] = 1  # added
        forest.networks[network_id][tree_2].data[tmp_tree_connections.connection_solutions[i].vessel_2_id, 24] = 0 # added
        for vdx in range(1,tmp_tree_connections.connection_solutions[i].vessels[1].shape[0]):
            tmp_tree_connections.connection_solutions[i].vessels[1][vdx,-1] = next_vessel_index_2
            tmp_tree_connections.connection_solutions[i].vessels[1][vdx,17] = parent_vessel_index
            tmp_tree_connections.connection_solutions[i].vessels[1][vdx,18] = previous_node_index
            tmp_tree_connections.connection_solutions[i].vessels[1][vdx,19] = next_node_index
            tmp_tree_connections.connection_solutions[i].vessels[1][vdx,21] = tmp_tree_connections.connection_solutions[i].vessels[1][vdx-1,21]
            tmp_tree_connections.connection_solutions[i].vessels[1][vdx,22] = tmp_tree_connections.connection_solutions[i].vessels[1][vdx-1,22]
            tmp_tree_connections.connection_solutions[i].vessels[1][vdx,26] = tmp_tree_connections.connection_solutions[i].vessels[1][vdx-1,26]+1
            tmp_tree_connections.connection_solutions[i].vessels[1][vdx,28] = tmp_tree_connections.connection_solutions[i].vessels[1][vdx-1,28]
            tmp_tree_connections.connection_solutions[i].vessels[1][vdx-1,15] = next_vessel_index_2
            tmp_tree_connections.connection_solutions[i].vessels[1][vdx-1,23] = 1
            tmp_tree_connections.connection_solutions[i].vessels[1][vdx-1,24] = 0
            parent_vessel_index = next_vessel_index_2
            next_vessel_index_2 += 1
            previous_node_index = next_node_index
            next_node_index += 1
            count_2 += 1
        tree_1_connections.append(tmp_tree_connections.connection_solutions[i].vessels[0][1:,:])
        tree_2_connections.append(tmp_tree_connections.connection_solutions[i].vessels[1][1:,:])
    for i in range(1,len(tmp_tree_connections.connection_solutions)):
        tree_1_connections[0] = np.vstack((tree_1_connections[0],tree_1_connections[i]))
        tree_2_connections[0] = np.vstack((tree_2_connections[0],tree_2_connections[i]))
    tree_1_connections = tree_1_connections[0]
    tree_2_connections = tree_2_connections[0]
    all_connections = [[tree_1_connections,tree_2_connections]]
    return all_connections, None, None

def link_v2(forest,network_id,tree_idx,tree_jdx,radius_buffer=0):
    """
    Method for connecting vascular trees together

    Parameters
    ----------
                forest : svcco.forest object
                       the forest object containing at least two trees to be
                       connected
                network_id : int
                       the index of the network containing two trees that will be
                       connected
                tree_idx : int
                       the index of the first tree in the network for connecting
                tree_jdx : int
                       the index of the second tree in the network for connecting
                radius_buffer : float
                       the minimum distance between the connecting vessels and any
                       other vessel within the network

    Returns
    -------
                connections : list of connection objects
                        for each connection between terminals in a network a
                        connection object is generated and solved
    """
    number_of_connections = forest.assignments[network_id][tree_idx].shape[0]
    connections = []
    assign_idx = forest.assignments[network_id][tree_idx]
    assign_jdx = forest.assignments[network_id][tree_jdx]
    print('Linking Network {}, Tree {} -> Tree {}'.format(network_id,tree_idx,tree_jdx))
    for cdx in trange(number_of_connections):
        conn = connection(forest,network_id,tree_idx,tree_jdx,cdx,radius_buffer)
        for pconn in connections:
            conn.add_collision_vessels(pconn)
        conn.solve()
        conn.build_vessels()
        connections.append(conn)
    tree_idx_connections = []
    tree_jdx_connections = []
    count_idx = 0
    count_jdx = 0
    for cdx in trange(number_of_connections):
        # Replace the current terminal vessel with the first vessel of the connection
        forest.networks[network_id][tree_idx].data[assign_idx[cdx],0:3]   = connections[cdx].vessels[0][0,0:3]
        forest.networks[network_id][tree_idx].data[assign_idx[cdx],3:6]   = connections[cdx].vessels[0][0,3:6]
        forest.networks[network_id][tree_idx].data[assign_idx[cdx],12:15] = connections[cdx].vessels[0][0,12:15]
        forest.networks[network_id][tree_idx].data[assign_idx[cdx],20]    = connections[cdx].vessels[0][0,20]
        connections[cdx].vessels[0][0,21] = forest.networks[network_id][tree_idx].data[assign_idx[cdx],21]
        connections[cdx].vessels[0][0,22] = forest.networks[network_id][tree_idx].data[assign_idx[cdx],22]
        connections[cdx].vessels[0][0,26] = forest.networks[network_id][tree_idx].data[assign_idx[cdx],26]
        connections[cdx].vessels[0][0,28] = forest.networks[network_id][tree_idx].data[assign_idx[cdx],28]
        parent_vessel_index = int(assign_idx[cdx])
        next_vessel_index   = int(max(forest.networks[network_id][tree_idx].data[:,-1]))+1+count_idx
        previous_node_index = int(forest.networks[network_id][tree_idx].data[assign_idx[cdx],19])
        next_node_index     = int(max(forest.networks[network_id][tree_idx].data[:,19]))+1
        for vdx in range(1,connections[cdx].vessels[0].shape[0]):
            connections[cdx].vessels[0][vdx,-1] = next_vessel_index
            connections[cdx].vessels[0][vdx,17] = parent_vessel_index
            connections[cdx].vessels[0][vdx,18] = previous_node_index
            connections[cdx].vessels[0][vdx,19] = next_node_index
            connections[cdx].vessels[0][vdx,21] = connections[cdx].vessels[0][vdx-1,21]
            connections[cdx].vessels[0][vdx,22] = connections[cdx].vessels[0][vdx-1,22]
            connections[cdx].vessels[0][vdx,26] = connections[cdx].vessels[0][vdx-1,26]+1
            connections[cdx].vessels[0][vdx,28] = connections[cdx].vessels[0][vdx-1,28]
            connections[cdx].vessels[0][vdx-1,15] = next_vessel_index
            connections[cdx].vessels[0][vdx-1,23] = 1
            connections[cdx].vessels[0][vdx-1,24] = 0
            #forest.networks[network_id][tree_idx].data = np.vstack((forest.networks[network_id][tree_idx].data,connections[cdx].vessels[0][vdx,:].reshape(1,-1)))
            parent_vessel_index = next_vessel_index
            next_vessel_index += 1
            previous_node_index = next_node_index
            next_node_index += 1
            count_idx += 1
        tree_idx_connections.append(connections[cdx].vessels[0][1:,:])
        # Replace the current terminal vessel with the first vessel of the connection
        # for tree jdx
        forest.networks[network_id][tree_jdx].data[assign_jdx[cdx],0:3]   = connections[cdx].vessels[1][0,0:3]
        forest.networks[network_id][tree_jdx].data[assign_jdx[cdx],3:6]   = connections[cdx].vessels[1][0,3:6]
        forest.networks[network_id][tree_jdx].data[assign_jdx[cdx],12:15] = connections[cdx].vessels[1][0,12:15]
        forest.networks[network_id][tree_jdx].data[assign_jdx[cdx],20]    = connections[cdx].vessels[1][0,20]
        connections[cdx].vessels[1][0,21] = forest.networks[network_id][tree_jdx].data[assign_jdx[cdx],21]
        connections[cdx].vessels[1][0,22] = forest.networks[network_id][tree_jdx].data[assign_jdx[cdx],22]
        connections[cdx].vessels[1][0,26] = forest.networks[network_id][tree_jdx].data[assign_jdx[cdx],26]
        connections[cdx].vessels[1][0,28] = forest.networks[network_id][tree_jdx].data[assign_jdx[cdx],28]
        parent_vessel_index = int(assign_jdx[cdx])
        next_vessel_index   = int(max(forest.networks[network_id][tree_jdx].data[:,-1]))+1+count_jdx
        previous_node_index = int(forest.networks[network_id][tree_jdx].data[assign_jdx[cdx],19])
        next_node_index     = int(max(forest.networks[network_id][tree_jdx].data[:,19]))+1
        for vdx in range(1,connections[cdx].vessels[1].shape[0]):
            connections[cdx].vessels[1][vdx,-1] = next_vessel_index
            connections[cdx].vessels[1][vdx,17] = parent_vessel_index
            connections[cdx].vessels[1][vdx,18] = previous_node_index
            connections[cdx].vessels[1][vdx,19] = next_node_index
            connections[cdx].vessels[1][vdx,21] = connections[cdx].vessels[1][vdx-1,21]
            connections[cdx].vessels[1][vdx,22] = connections[cdx].vessels[1][vdx-1,22]
            connections[cdx].vessels[1][vdx,26] = connections[cdx].vessels[1][vdx-1,26]+1
            connections[cdx].vessels[1][vdx,28] = connections[cdx].vessels[1][vdx-1,28]
            connections[cdx].vessels[1][vdx-1,15] = next_vessel_index
            connections[cdx].vessels[1][vdx-1,23] = 1
            connections[cdx].vessels[1][vdx-1,24] = 0
            #forest.networks[network_id][tree_jdx].data = np.vstack((forest.networks[network_id][tree_jdx].data,connections[cdx].vessels[1][vdx,:].reshape(1,-1)))
            parent_vessel_index = next_vessel_index
            next_vessel_index += 1
            previous_node_index = next_node_index
            next_node_index += 1
            count_jdx += 1
        tree_jdx_connections.append(connections[cdx].vessels[1][1:,:])
    for cdx in range(1,len(tree_idx_connections)):
        tree_idx_connections[0] = np.vstack((tree_idx_connections[0],tree_idx_connections[cdx]))
    for cdx in range(1,len(tree_jdx_connections)):
        tree_jdx_connections[0] = np.vstack((tree_jdx_connections[0],tree_jdx_connections[cdx]))
    tree_idx_connections = tree_idx_connections[0]
    tree_jdx_connections = tree_jdx_connections[0]
    all_connections = [[tree_idx_connections,tree_jdx_connections]]
    return all_connections, None, None

def link(forest,radius_buffer=0):
    all_connections = []
    all_network_vessels = np.zeros((1,7))
    for network in range(forest.number_of_networks):
        all_network_vessels = np.zeros((1,7))
        for tree in range(forest.trees_per_network[network]):
            tree_tmp_vessels = np.zeros((forest.networks[network][tree].data.shape[0],7))
            tree_tmp_vessels[:,0:6] = forest.networks[network][tree].data[:,0:6]
            tree_tmp_vessels[:,6]   = forest.networks[network][tree].data[:,21]
            all_network_vessels = np.vstack((all_network_vessels,tree_tmp_vessels))
    all_network_vessels = all_network_vessels[1:,:]
    all_network_mid_points = (all_network_vessels[:,0:3]+all_network_vessels[:,3:6])/2
    for network in range(forest.number_of_networks):
        network_connections = []
        for tree in range(forest.trees_per_network[network]):
            if tree == 1:
                continue
            if tree == 0:
                tree_connections_1 = []
                tree_connections_2 = []
                tree_edge_count_1 = max(forest.networks[network][tree].data[:,-1]) + 1
                tree_edge_count_2 = max(forest.networks[network][1].data[:,-1]) + 1
                assignment_1 = forest.assignments[network][tree]
                assignment_2 = forest.assignments[network][1]
            else:
                tree_connections = []
                tree_edge_count = max(forest.networks[network][tree].data[:,-1]) + 1
                assignment = forest.assignments[network][tree]
            print('Linking Network {}, Tree {}'.format(network,tree))
            for edge in trange(len(forest.assignments[network][0])):
                if tree == 0:
                    R        = forest.networks[network][tree].data[assignment_1[edge],21]
                    P1       = forest.networks[network][tree].data[assignment_1[edge],0:3]
                    P2       = forest.networks[network][tree].data[assignment_1[edge],3:6]
                    ignore   = set(np.argwhere(np.all(all_network_mid_points[3:6]==P2)).flatten().tolist())
                    R_tree_1 = forest.networks[network][1].data[assignment_2[edge],21]
                    P3       = forest.networks[network][1].data[assignment_2[edge],0:3]
                    P4       = forest.networks[network][1].data[assignment_2[edge],3:6]
                    ignore_other = set(np.argwhere(np.all(all_network_mid_points[3:6]==P4)).flatten().tolist())
                    ignore = ignore.union(ignore_other)
                else:
                    R        = forest.networks[network][tree].data[assignment[edge],21]
                    P1       = forest.networks[network][tree].data[assignment[edge],0:3]
                    P2       = forest.networks[network][tree].data[assignment[edge],3:6]
                    ignore   = set(np.argwhere(np.all(all_network_mid_points[3:6]==P2)).flatten().tolist())
                    P4       = forest.connections[network][0][edge]
                    V        = (P4 - P2)/np.linalg.norm(P4-P2)
                    P3       = P4 + V*0.25
                dist_to_mid = np.linalg.norm(P4-P2)
                dists       = np.linalg.norm(all_network_mid_points - P4)
                collision_vessels_idx = set(np.argwhere(dists<dist_to_mid).flatten().tolist())
                if len(collision_vessels_idx) == 0:
                    collision_vessels = None
                else:
                    collision_vessels_idx = np.array(list(collision_vessels_idx - ignore))
                    collision_vessels = all_network_vessels[collision_vessels_idx,:]
                if tree == 0:
                    bounds = np.array([forest.boundary.x_range,forest.boundary.y_range,forest.boundary.z_range])
                    link_pts = get_optimum_link_points(P1,P2,P3,P4,R,collision_vessels,bounds=bounds,radius_buffer=radius_buffer)
                    num_pts  = link_pts.shape[0]
                    mid_num  = int(num_pts // 2)
                    link_pts_1 = link_pts[1:mid_num+1]
                    link_pts_2 = link_pts[mid_num:-1]
                    link_pts_1 = np.vstack((P2,link_pts_1))
                    link_pts_2 = np.vstack((link_pts_2,P4))
                    link_pts_2 = np.flip(link_pts_2,axis=0)
                    forest.connections[network][0][edge] = link_pts[mid_num,:]
                    for i in range(link_pts_1.shape[0]-1):
                        tmp = np.zeros(forest.networks[network][tree].data.shape[1])
                        tmp[0:3] = link_pts_1[i,:]
                        tmp[3:6] = link_pts_1[i+1,:]
                        tmp[21]  = forest.networks[network][tree].data[assignment_1[edge],21]
                        tmp[20]  = np.linalg.norm(link_pts_1[i+1]-link_pts_1[i])
                        tmp[12:15] = (link_pts_1[i+1]-link_pts_1[i])/tmp[20]
                        tmp[26] = forest.networks[network][tree].data[assignment_1[edge],26]+i+1
                        tmp[22] = forest.networks[network][tree].data[assignment_1[edge],22]
                        tmp[-1] = tree_edge_count_1
                        tmp[15] = -1
                        tmp[16] = -1
                        tree_edge_count_1 += 1
                        if i == 0:
                            tmp[17] = forest.networks[network][tree].data[assignment_1[edge],-1]
                        elif i > 0:
                            tree_connections_1[-1][15] = tmp[-1]
                            tree_connections_1[-1][16] = -1
                            tmp[17] = tree_connections_1[-1][-1]
                        tmp_mid = (tmp[0:3]+tmp[3:6])/2
                        all_network_mid_points = np.vstack((all_network_mid_points,tmp_mid.reshape(1,3)))
                        tree_connections_1.append(tmp.tolist())
                    for i in range(link_pts_2.shape[0]-1):
                        tmp = np.zeros(forest.networks[network][tree].data.shape[1])
                        tmp[0:3] = link_pts_2[i,:]
                        tmp[3:6] = link_pts_2[i+1,:]
                        tmp[21]  = forest.networks[network][tree].data[assignment_2[edge],21]
                        tmp[20]  = np.linalg.norm(link_pts_2[i+1]-link_pts_2[i])
                        tmp[12:15] = (link_pts_2[i+1]-link_pts_2[i])/tmp[20]
                        tmp[26] = forest.networks[network][1].data[assignment_2[edge],26]+i+1
                        tmp[22] = forest.networks[network][1].data[assignment_2[edge],22]
                        tmp[-1] = tree_edge_count_2
                        tmp[15] = -1
                        tmp[16] = -1
                        tree_edge_count_2 += 1
                        if i == 0:
                            tmp[17] = forest.networks[network][1].data[assignment_2[edge],-1]
                        elif i > 0:
                            tree_connections_2[-1][15] = tmp[-1]
                            tree_connections_2[-1][16] = -1
                            tmp[17] = tree_connections_2[-1][-1]
                        tmp_mid = (tmp[0:3]+tmp[3:6])/2
                        all_network_mid_points = np.vstack((all_network_mid_points,tmp_mid.reshape(1,3)))
                        tree_connections_2.append(tmp.tolist())
                else:
                    bounds = np.array([forest.boundary.x_range,forest.boundary.y_range,forest.boundary.z_range])
                    link_pts = get_optimum_link_points(P1,P2,P3,P4,R,collision_vessels,bounds=bounds,radius_buffer=radius_buffer)
                    link_pts = link_pts[1:-1,:]
                    link_pts = np.vstack((P2,link_pts,P4))
                    for i in range(link_pts.shape[0]-1):
                        tmp = np.zeros(forest.networks[network][tree].data.shape[1])
                        tmp[0:3] = link_pts[i,:]
                        tmp[3:6] = link_pts[i+1,:]
                        tmp[21]  = forest.networks[network][tree].data[assignment[edge],21]
                        tmp[20]  = np.linalg.norm(link_pts[i+1]-link_pts[i])
                        tmp[12:15] = (link_pts[i+1]-link_pts[i])/tmp[20]
                        tmp[26] = forest.networks[network][tree].data[assignment[edge],26]+i+1
                        tmp[22] = forest.networks[network][tree].data[assignment[edge],22]
                        tmp[-1] = tree_edge_count
                        tmp[15] = -1
                        tmp[16] = -1
                        tree_edge_count += 1
                        if i == 0:
                            tmp[17] = forest.networks[network][tree].data[assignment[edge],-1]
                        elif i > 0:
                            tree_connections[-1][15] = tmp[-1]
                            tree_connections[-1][16] = -1
                            tmp[17] = tree_connections[-1][-1]
                        tmp_mid = (tmp[0:3]+tmp[3:6])/2
                        all_network_mid_points = np.vstack((all_network_mid_points,tmp_mid.reshape(1,3)))
                        tree_connections.append(tmp.tolist())
            if tree == 0:
                network_connections.append(np.array(tree_connections_1))
                network_connections.append(np.array(tree_connections_2))
            else:
                network_connections.append(np.array(tree_connections))
        all_connections.append(network_connections)
    return all_connections, None, None

def smooth_original(forest,curve_sample_size_min=5,curve_sample_size_max=21,curve_degree=3):
    #FOREST_CONNECT_COPY = deepcopy(forest)
    connections = []
    for network in range(forest.number_of_networks):
        if not np.any(forest.connections[network][0] is None):
            network_connections = []
            for tree in range(forest.trees_per_network[network]):
                tree_assignment = forest.assignments[network][tree]
                idx = max(forest.networks[network][tree].data[:,-1]) + 1
                #tree_connection_edges = np.zeros((len(forest.connections[network][0])*(curve_sample_size-1),forest.networks[network][tree].data.shape[1]))
                tree_connection_edges = []
                for edge in range(len(forest.connections[network][0])):
                    #print('Tree {}: Edge: {}'.format(tree,edge))
                    success = False
                    samples = curve_sample_size_min
                    # Check the angle formed
                    vec_1 = forest.networks[network][tree].data[tree_assignment[edge],0:3]-forest.networks[network][tree].data[tree_assignment[edge],3:6]
                    vec_2 = forest.connections[network][0][edge]-forest.networks[network][tree].data[tree_assignment[edge],3:6]
                    vec_1 = vec_1/np.linalg.norm(vec_1)
                    vec_2 = vec_2/np.linalg.norm(vec_2)
                    P0    = (forest.networks[network][tree].data[tree_assignment[edge],0:3]+forest.networks[network][tree].data[tree_assignment[edge],3:6])/2
                    P1    = forest.networks[network][tree].data[tree_assignment[edge],3:6]
                    ctr_0 = ((forest.networks[network][tree].data[tree_assignment[edge],0:3]*(3/4)+forest.networks[network][tree].data[tree_assignment[edge],3:6]*(1/4))).flatten().tolist()
                    ctr_1 = ((forest.networks[network][tree].data[tree_assignment[edge],0:3]+forest.networks[network][tree].data[tree_assignment[edge],3:6])/2).flatten().tolist()
                    if (np.arccos(np.dot(vec_1,vec_2))/np.pi)*180 > 120:
                        ctr_2 = forest.networks[network][tree].data[tree_assignment[edge],3:6].flatten().tolist()
                        ctr_3 = forest.connections[network][0][edge].tolist()
                    else:
                        A = 140
                        t = ((np.cos(np.pi*(A/180))-P0[0]*vec_2[0]-P0[1]*vec_2[1]-P0[2]*vec_2[2]+P1[0]*vec_2[0]+P1[1]*vec_2[1]+P1[2]*vec_2[2])/(vec_2[0]**2+vec_2[1]**2+vec_2[2]**2))
                        t = min(t,0.3)
                        conn_len = np.linalg.norm(forest.connections[network][0][edge]-forest.networks[network][tree].data[tree_assignment[edge],3:6])/2
                    #    vec_12_average = (vec_1+vec_2)/2
                        adj_0 = (P1-t*(conn_len)*vec_2)
                        if conn_len < forest.networks[network][tree].data[tree_assignment[edge],21]:
                            print("WARNING: Sharp Corner, lofting may fail")
                    #    vec_adj = adj_0 - P0/np.linalg.norm(adj_0 - P0)
                    #    vec_diff = vec_adj - vec_12_average
                    #    vec_diff = np.argwhere(vec_diff != 0).flatten()
                    #    vec_idx  = vec_diff[:2]
                    #    #AA = np.array([[vec_adj[vec_idx[0]],vec_12_average[vec_idx[0]]],
                    #    #               [vec_adj[vec_idx[1]],vec_12_average[vec_idx[1]]]])
                    #    #BB = np.array([[P0[vec_idx[0]]-P1[vec_idx[0]]],
                    #    #               [P0[vec_idx[1]]-P1[vec_idx[1]]]])
                    #    #AA_inv = np.linalg.inv(AA)
                    #    #sol = BB @ AA
                    #    sol  = (P0-P1)/(vec_adj - vec_12_average)
                    #    tt = sol[0]
                        ctr_2 = ((P0 + adj_0)/2).flatten().tolist()
                        ctr_3 = forest.connections[network][0][edge].tolist()
                    curve = BSpline.Curve()
                    curve.degree = curve_degree
                    #if (np.arccos(np.dot(vec_1,vec_2))/np.pi)*180 > 90:
                    curve.ctrlpts = [ctr_1,ctr_2,ctr_3]
                    #else:
                    #curve.ctrlpts = [ctr_0,ctr_1,ctr_2,ctr_3]
                    curve.knotvector = utilities.generate_knot_vector(curve.degree,len(curve.ctrlpts))
                    while samples <= curve_sample_size_max and not success:
                        curve.sample_size = samples
                        curve.evaluate()
                        pts = curve.evalpts
                        d0  = forest.networks[network][tree].data[tree_assignment[edge],12:15]
                        for p in range(len(pts)-1):
                            p0 = np.array(pts[p])
                            p1 = np.array(pts[p+1])
                            l  = np.linalg.norm(p1-p0)
                            d1 = (p1-p0)/l
                            A  = (abs(np.arccos(np.dot(-d0,d1)))/np.pi)*180
                            if A < 90:
                                success = False
                                samples += 2
                                break
                            else:
                                success = True
                            d0 = d1
                    if samples > curve_sample_size_max:
                        samples -= 2
                    for i in range(samples-1):
                        tmp = np.zeros(forest.networks[network][tree].data.shape[1])
                        tmp[0:3] = np.array(pts[i])
                        tmp[3:6] = np.array(pts[i+1])
                        tmp[21]  = forest.networks[network][tree].data[tree_assignment[edge],21]
                        tmp[20]  = np.linalg.norm(np.array(pts[i+1])-np.array(pts[i]))
                        tmp[12:15] = (np.array(pts[i+1])-np.array(pts[i]))/tmp[20]
                        tmp[26] = forest.networks[network][tree].data[tree_assignment[edge],26]+i+1
                        tmp[22] = forest.networks[network][tree].data[tree_assignment[edge],22]
                        tmp[-1] = idx
                        tmp[15] = -1
                        tmp[16] = -1
                        idx += 1
                        if i == 0:
                            tmp[17] = forest.networks[network][tree].data[tree_assignment[edge],-1]
                        elif i > 0:
                            tree_connection_edges[-1][15] = tmp[-1]
                            tree_connection_edges[-1][16] = -1
                            tmp[17] = tree_connection_edges[-1][-1]
                        #else:
                        #    tree_connection_edges[-1][15] = -1
                        #    tree_connection_edges[-1][16] = -1
                        #    tmp[17] = tree_connection_edges[-1][-1]
                        tree_connection_edges.append(tmp.tolist())
                network_connections.append(np.array(tree_connection_edges))
            connections.append(network_connections)
        else:
            connections.append([None])
    return connections, None

def smooth(forest,curve_sample_size_min=5,curve_sample_size_max=21,curve_degree=3):
    #FOREST_CONNECT_COPY = deepcopy(forest)
    connections = []
    splines     = []
    for network in range(forest.number_of_networks):
        if not np.any(forest.connections[network][0] is None):
            network_connections = []
            network_splines     = []
            for tree in range(forest.trees_per_network[network]):
                tree_assignment = forest.assignments[network][tree]
                idx = max(forest.networks[network][tree].data[:,-1]) + 1
                #tree_connection_edges = np.zeros((len(forest.connections[network][0])*(curve_sample_size-1),forest.networks[network][tree].data.shape[1]))
                tree_connection_edges = []
                for edge in range(len(forest.connections[network][0])):
                    #print('Tree {}: Edge: {}'.format(tree,edge))
                    success = False
                    samples = curve_sample_size_min
                    # Check the angle formed
                    vec_1 = forest.networks[network][tree].data[tree_assignment[edge],0:3]-forest.networks[network][tree].data[tree_assignment[edge],3:6]
                    vec_2 = forest.connections[network][0][edge]-forest.networks[network][tree].data[tree_assignment[edge],3:6]
                    vec_1 = vec_1/np.linalg.norm(vec_1)
                    vec_2 = vec_2/np.linalg.norm(vec_2)
                    print((np.arccos(np.dot(vec_1,vec_2))/np.pi)*180)
                    P0    = (forest.networks[network][tree].data[tree_assignment[edge],0:3]+forest.networks[network][tree].data[tree_assignment[edge],3:6])/2
                    P1    = forest.networks[network][tree].data[tree_assignment[edge],3:6]
                    ctr_0 = ((forest.networks[network][tree].data[tree_assignment[edge],0:3]*(3/4)+forest.networks[network][tree].data[tree_assignment[edge],3:6]*(1/4))).flatten().tolist()
                    #ctr_1 = ((forest.networks[network][tree].data[tree_assignment[edge],0:3]+forest.networks[network][tree].data[tree_assignment[edge],3:6])/2).flatten().tolist()
                    ctr_1 = P0.flatten().tolist()
                    if False:
                        #if (np.arccos(np.dot(vec_1,vec_2))/np.pi)*180 > 0:
                        ctr_2 = forest.networks[network][tree].data[tree_assignment[edge],3:6].flatten().tolist()
                        ctr_3 = forest.connections[network][0][edge].tolist()
                    else:
                        A = 120
                        t = ((np.cos(np.pi*(A/180))-P0[0]*vec_2[0]-P0[1]*vec_2[1]-P0[2]*vec_2[2]+P1[0]*vec_2[0]+P1[1]*vec_2[1]+P1[2]*vec_2[2])/(vec_2[0]**2+vec_2[1]**2+vec_2[2]**2))
                        t = min(t,0.3)
                        conn_len = np.linalg.norm(forest.connections[network][0][edge]-forest.networks[network][tree].data[tree_assignment[edge],3:6])/2
                    #    vec_12_average = (vec_1+vec_2)/2
                        adj_0 = (P1-t*(conn_len)*vec_2)
                        if conn_len < forest.networks[network][tree].data[tree_assignment[edge],21]:
                            print("WARNING: Sharp Corner, lofting may fail")
                    #    vec_adj = adj_0 - P0/np.linalg.norm(adj_0 - P0)
                    #    vec_diff = vec_adj - vec_12_average
                    #    vec_diff = np.argwhere(vec_diff != 0).flatten()
                    #    vec_idx  = vec_diff[:2]
                    #    #AA = np.array([[vec_adj[vec_idx[0]],vec_12_average[vec_idx[0]]],
                    #    #               [vec_adj[vec_idx[1]],vec_12_average[vec_idx[1]]]])
                    #    #BB = np.array([[P0[vec_idx[0]]-P1[vec_idx[0]]],
                    #    #               [P0[vec_idx[1]]-P1[vec_idx[1]]]])
                    #    #AA_inv = np.linalg.inv(AA)
                    #    #sol = BB @ AA
                    #    sol  = (P0-P1)/(vec_adj - vec_12_average)
                    #    tt = sol[0]
                        ctr_2 = ((P0+P1)/2).flatten().tolist()
                        ctr_3 = ((P1+adj_0)/2).flatten().tolist()
                        ctr_4 = forest.connections[network][0][edge].tolist()
                    curve = BSpline.Curve()
                    curve.degree = curve_degree
                    #if (np.arccos(np.dot(vec_1,vec_2))/np.pi)*180 > 90:
                    curve.ctrlpts = [ctr_1,ctr_2,ctr_3,ctr_4]
                    #else:
                    #curve.ctrlpts = [ctr_0,ctr_1,ctr_2,ctr_3]
                    curve.knotvector = utilities.generate_knot_vector(curve.degree,len(curve.ctrlpts))
                    network_splines.append(deepcopy(curve))
                    while samples <= curve_sample_size_max and not success:
                        curve.sample_size = samples
                        curve.evaluate()
                        pts = curve.evalpts
                        d0  = forest.networks[network][tree].data[tree_assignment[edge],12:15]
                        for p in range(len(pts)-1):
                            p0 = np.array(pts[p])
                            p1 = np.array(pts[p+1])
                            l  = np.linalg.norm(p1-p0)
                            d1 = (p1-p0)/l
                            A  = (abs(np.arccos(np.dot(-d0,d1)))/np.pi)*180
                            if A < 90:
                                success = False
                                samples += 2
                                break
                            else:
                                success = True
                            d0 = d1
                    if samples > curve_sample_size_max:
                        samples -= 2
                    for i in range(samples-1):
                        tmp = np.zeros(forest.networks[network][tree].data.shape[1])
                        tmp[0:3] = np.array(pts[i])
                        tmp[3:6] = np.array(pts[i+1])
                        tmp[21]  = forest.networks[network][tree].data[tree_assignment[edge],21]
                        tmp[20]  = np.linalg.norm(np.array(pts[i+1])-np.array(pts[i]))
                        tmp[12:15] = (np.array(pts[i+1])-np.array(pts[i]))/tmp[20]
                        tmp[26] = forest.networks[network][tree].data[tree_assignment[edge],26]+i+1
                        tmp[22] = forest.networks[network][tree].data[tree_assignment[edge],22]
                        tmp[-1] = idx
                        tmp[15] = -1
                        tmp[16] = -1
                        idx += 1
                        if i == 0:
                            tmp[17] = forest.networks[network][tree].data[tree_assignment[edge],-1]
                        elif i > 0:
                            tree_connection_edges[-1][15] = tmp[-1]
                            tree_connection_edges[-1][16] = -1
                            tmp[17] = tree_connection_edges[-1][-1]
                        #else:
                        #    tree_connection_edges[-1][15] = -1
                        #    tree_connection_edges[-1][16] = -1
                        #    tmp[17] = tree_connection_edges[-1][-1]
                        tree_connection_edges.append(tmp.tolist())

                network_connections.append(np.array(tree_connection_edges))
            connections.append(network_connections)
            splines.append(network_splines)
        else:
            connections.append([None])
    return connections, None, splines

def smooth_new(forest,max_angle=45):
    connections = []
    for network in range(forest.number_of_networks):
        if not np.any(forest.connections[network][0] is None):
            network_connections = []
            for tree in range(forest.trees_per_network[network]):
                idx = max(forest.networks[network][tree].data[:,-1]) + 1
                tree_connection_edges = []
                for edge in range(len(forest.connections[network][0])):
                    edge_1   = forest.assignments[network][0][edge]
                    edge_2   = forest.assignemnts[network][1][edge]
                    match_vector = forest.networks[network][0].data[edge_1,3:6] - forest.networks[network][1].data[edge_2,3:6]
                    match_vector = match_vector/np.linalg.norm(match_vector)
                    vector_1 = forest.networks[network][0].data[edge_1,12:15]
                    vector_2 = forest.networks[network][1].data[edge_2,12:15]

    return connections, None



#!/usr/bin/env python
import vtk
import os
import numpy as np

from vtk.util.numpy_support import vtk_to_numpy as v2n
from tqdm import tqdm

from get_bc_integrals import get_res_names
from get_database import input_args
from vtk_functions import read_geo, write_geo, calculator, cut_plane, connectivity, get_points_cells, clean, Integration

import matplotlib.pyplot as plt


def slice_vessel(inp_3d, origin, normal):
    """
    Slice 3d geometry at certain plane
    Args:
        inp_1d: vtk InputConnection for 1d centerline
        inp_3d: vtk InputConnection for 3d volume model
        origin: plane origin
        normal: plane normal

    Returns:
        Integration object
    """
    # cut 3d geometry
    cut_3d = cut_plane(inp_3d, origin, normal)

    # extract region closest to centerline
    con = connectivity(cut_3d, origin)

    return con


def get_integral(inp_3d, origin, normal):
    """
    Slice simulation at certain plane and integrate
    Args:
        inp_1d: vtk InputConnection for 1d centerline
        inp_3d: vtk InputConnection for 3d volume model
        origin: plane origin
        normal: plane normal

    Returns:
        Integration object
    """
    # slice vessel at given location
    inp = slice_vessel(inp_3d, origin, normal)

    # recursively add calculators for normal velocities
    for v in get_res_names(inp_3d, 'velocity'):
        fun = '(iHat*'+repr(normal[0])+'+jHat*'+repr(normal[1])+'+kHat*'+repr(normal[2])+').' + v
        inp = calculator(inp, fun, [v], 'normal_' + v)

    return Integration(inp)


def extract_results(fpath_1d, fpath_3d, fpath_out, only_caps=False):
    """
    Extract 3d results at 1d model nodes (integrate over cross-section)
    Args:
        fpath_1d: path to 1d model
        fpath_3d: path to 3d simulation results
        fpath_out: output path
        only_caps: extract solution only at caps, not in interior (much faster)

    Returns:
        res: dictionary of results in all branches, in all segments for all result arrays
    """
    # read 1d and 3d model
    reader_1d = read_geo(fpath_1d).GetOutput()
    reader_3d = read_geo(fpath_3d).GetOutput()

    # get all result array names
    res_names = get_res_names(reader_3d, ['pressure', 'velocity'])

    # get point and normals from centerline
    points = v2n(reader_1d.GetPoints().GetData())
    normals = v2n(reader_1d.GetPointData().GetArray('CenterlineSectionNormal'))
    gid = v2n(reader_1d.GetPointData().GetArray('GlobalNodeId'))

    # initialize output
    for name in res_names + ['area']:
        array = vtk.vtkDoubleArray()
        array.SetName(name)
        array.SetNumberOfValues(reader_1d.GetNumberOfPoints())
        array.Fill(0)
        reader_1d.GetPointData().AddArray(array)

    # move points on caps slightly to ensure nice integration
    ids = vtk.vtkIdList()
    eps_norm = 1.0e-3

    # integrate results on all points of intergration cells
    for i in tqdm(range(reader_1d.GetNumberOfPoints())):
        # check if point is cap
        reader_1d.GetPointCells(i, ids)
        if ids.GetNumberOfIds() == 1:
            if gid[i] == 0:
                # inlet
                points[i] += eps_norm * normals[i]
            else:
                # outlets
                points[i] -= eps_norm * normals[i]
        else:
            if only_caps:
                continue

        # create integration object (slice geometry at point/normal)
        try:
            integral = get_integral(reader_3d, points[i], normals[i])
        except Exception:
            continue

        # integrate all output arrays
        for name in res_names:
            reader_1d.GetPointData().GetArray(name).SetValue(i, integral.evaluate(name))
        reader_1d.GetPointData().GetArray('area').SetValue(i, integral.area())

    write_geo(fpath_out, reader_1d)


def main(db, geometries):
    """
    Loop all geometries
    """
    for geo in geometries:
        print('Running geometry ' + geo)

        fpath_1d = db.get_centerline_path(geo)
        fpath_3d = db.get_volume(geo)
        fpath_out = db.get_3d_flow(geo)

        # if os.path.exists(db.get_3d_flow_path_oned_vtp(geo)):
        #     print('  Already exists. Skipping...')
        #     continue

        if not os.path.exists(fpath_1d) or not os.path.exists(fpath_3d):
            continue

        # extract 3d results integrated over cross-section
        try:
            extract_results(fpath_1d, fpath_3d, fpath_out)
        except Exception as e:
            print(e)
            continue


if __name__ == '__main__':
    descr = 'Extract 3d-results at 1d-locations'
    d, g, _ = input_args(descr)
    main(d, g)

